from __future__ import annotations

import pytest

from jarvis_operas.catalog import get_catalog_declarations


def test_catalog_default_namespaces_do_not_include_interpolations() -> None:
    declarations = get_catalog_declarations()
    namespaces = {item.namespace for item in declarations}

    assert {"math", "stat", "helper"}.issubset(namespaces)
    assert "interp1" not in namespaces
    assert "dmdd" not in namespaces


def test_catalog_can_load_interpolation_namespaces_on_demand() -> None:
    declarations = get_catalog_declarations(
        namespaces=["interp1", "dmdd"],
        include_interpolations=True,
    )
    names = {item.full_name for item in declarations}

    assert "interp1.interp1_xy_flat" in names
    assert "dmdd.LZSI2024" in names


def test_catalog_rejects_unknown_namespace() -> None:
    with pytest.raises(KeyError):
        get_catalog_declarations(namespaces=["unknown_namespace"])



from __future__ import annotations

from .core import (
    OperaFunction,
    Operas,
    OperasRegistry,
    get_global_operas,
    get_global_operas_registry,
)

__all__ = [
    "OperaFunction",
    "OperasRegistry",
    "Operas",
    "get_global_operas_registry",
    "get_global_operas",
]



from __future__ import annotations

import json
import subprocess
import sys
from pathlib import Path


def test_generate_function_docs_script(tmp_path) -> None:
    repo_root = Path(__file__).resolve().parent.parent
    script = repo_root / "scripts" / "generate_function_docs.py"
    out_dir = tmp_path / "func_docs"

    subprocess.run(
        [sys.executable, str(script), "--out", str(out_dir)],
        cwd=repo_root,
        check=True,
        capture_output=True,
        text=True,
    )

    assert (out_dir / "index.md").is_file()
    assert (out_dir / "functions.json").is_file()

    index_payload = json.loads((out_dir / "functions.json").read_text(encoding="utf-8"))
    assert isinstance(index_payload, list)
    assert any(item["full_name"] == "math.add" for item in index_payload)
    assert any(item["full_name"] == "interp1.interp1_xy_flat" for item in index_payload)
    assert any(item["full_name"] == "dmdd.LZSI2024" for item in index_payload)

    math_add_page = out_dir / "math" / "add.md"
    assert math_add_page.is_file()
    body = math_add_page.read_text(encoding="utf-8")
    assert "Metadata (from `OperaFunction.metadata`)" in body
    assert "summary" in body

    assert (out_dir / "interp1" / "interp1_xy_flat.md").is_file()
    assert (out_dir / "dmdd" / "LZSI2024.md").is_file()



from __future__ import annotations

import jarvis_operas.api as api_mod
from jarvis_operas.core.registry import recommended_numpy_concurrency


def test_recommended_numpy_concurrency_scales_with_cpu_count() -> None:
    assert recommended_numpy_concurrency(cpu_count=1) == 3
    assert recommended_numpy_concurrency(cpu_count=4) == 12
    assert recommended_numpy_concurrency(cpu_count=0) == 3
    assert recommended_numpy_concurrency(cpu_count=-3) == 3


def test_get_global_operas_keeps_first_numpy_concurrency(
    tmp_path,
    monkeypatch,
) -> None:
    monkeypatch.setenv("JARVIS_OPERAS_CURVE_INDEX", str(tmp_path / "curve-index.json"))
    monkeypatch.setenv("JARVIS_OPERAS_PERSIST_FILE", str(tmp_path / "persist-store.json"))
    monkeypatch.setattr(api_mod, "_global_operas_registry", None)
    monkeypatch.setattr(api_mod, "_global_operas", None)

    first = api_mod.get_global_operas(numpy_concurrency=2)
    assert first.registry.get_numpy_concurrency() == 2

    second = api_mod.get_global_operas(numpy_concurrency=5)
    assert second is first
    assert second.registry.get_numpy_concurrency() == 2


def test_get_global_operas_default_concurrency_is_cpu_scaled(
    tmp_path,
    monkeypatch,
) -> None:
    monkeypatch.setenv("JARVIS_OPERAS_CURVE_INDEX", str(tmp_path / "curve-index.json"))
    monkeypatch.setenv("JARVIS_OPERAS_PERSIST_FILE", str(tmp_path / "persist-store.json"))
    monkeypatch.setattr(api_mod, "_global_operas_registry", None)
    monkeypatch.setattr(api_mod, "_global_operas", None)

    operas = api_mod.get_global_operas()
    concurrency = operas.registry.get_numpy_concurrency()

    assert concurrency >= 3
    assert concurrency % 3 == 0



from __future__ import annotations

import json
import os
import pickle
from pathlib import Path
from typing import Any, Mapping

from .io_utils import atomic_write_bytes

_CACHE_ROOT_ENV = "JARVIS_OPERAS_CURVE_CACHE_ROOT"
_INDEX_PATH_ENV = "JARVIS_OPERAS_CURVE_INDEX"
_INDEX_FILENAME = "index.json"
_INDEX_VERSION = 1
_PICKLE_PROTOCOL = 5


def get_curve_cache_root(cache_root: str | None = None) -> Path:
    if cache_root:
        return Path(cache_root).expanduser().resolve()
    env_value = os.getenv(_CACHE_ROOT_ENV)
    if env_value:
        return Path(env_value).expanduser().resolve()
    return (Path.home() / ".jarvis-operas" / "curve-cache").resolve()


def get_curve_index_path(
    *,
    cache_root: str | None = None,
    index_path: str | None = None,
) -> Path:
    if index_path:
        return Path(index_path).expanduser().resolve()
    env_value = os.getenv(_INDEX_PATH_ENV)
    if env_value:
        return Path(env_value).expanduser().resolve()
    return get_curve_cache_root(cache_root) / _INDEX_FILENAME


def _resolve_index_ref(path_ref: str, index_parent: Path) -> Path:
    path = Path(path_ref).expanduser()
    if path.is_absolute():
        return path.resolve()
    return (index_parent / path).resolve()


def _relative_or_absolute(path: Path, base: Path) -> str:
    try:
        return str(path.resolve().relative_to(base.resolve()))
    except ValueError:
        return str(path.resolve())


def _write_pickle(path: Path, value: Any) -> None:
    payload = pickle.dumps(value, protocol=_PICKLE_PROTOCOL)
    atomic_write_bytes(path, payload, fsync=True, temp_prefix=f"{path.name}.")


def _read_json_object(path: Path) -> dict[str, Any]:
    try:
        raw = json.loads(path.read_text(encoding="utf-8"))
    except Exception as exc:
        raise ValueError(f"invalid json file '{path}': {exc}") from exc
    if not isinstance(raw, Mapping):
        raise ValueError(f"json file '{path}' must contain an object")
    return dict(raw)


def _read_index(path: Path) -> dict[str, Any]:
    if not path.exists():
        return {"version": _INDEX_VERSION, "curves": {}}
    return _read_json_object(path)


def _write_json(path: Path, payload: Mapping[str, Any]) -> None:
    raw = json.dumps(payload, ensure_ascii=False, indent=2).encode("utf-8")
    atomic_write_bytes(path, raw, fsync=True, temp_prefix=f"{path.name}.")



from __future__ import annotations

from collections.abc import Sequence

from ...core.spec import OperaFunction
from .decls import MATH_BASIC_DECLARATIONS

DECLARATIONS: tuple[OperaFunction, ...] = MATH_BASIC_DECLARATIONS


def get_declarations() -> Sequence[OperaFunction]:
    return DECLARATIONS


__all__ = ["DECLARATIONS", "get_declarations"]



from __future__ import annotations

from ....core.spec import OperaFunction
from ..defs.basic import add_numpy, add_polars_expr, identity_numpy, identity_polars_expr


MATH_BASIC_DECLARATIONS: tuple[OperaFunction, ...] = (
    OperaFunction(
        namespace="math",
        name="add",
        arity=2,
        return_dtype=None,
        numpy_impl=add_numpy,
        polars_expr_impl=add_polars_expr,
        metadata={
            "category": "math",
            "summary": "Element-wise addition for scalar or array-like inputs.",
            "params": {
                "a": "Left operand (scalar/array-like).",
                "b": "Right operand (scalar/array-like).",
            },
            "return": "Sum of a and b with broadcast semantics from backend.",
            "backend_support": {
                "numpy": "native",
                "polars": "native_expr",
            },
            "examples": [
                "jopera call math.add --kwargs '{\"a\":1,\"b\":2}'",
            ],
            "since": "1.2.0",
            "tags": ["math", "arithmetic"],
        },
    ),
    OperaFunction(
        namespace="math",
        name="identity",
        arity=1,
        return_dtype=None,
        numpy_impl=identity_numpy,
        polars_expr_impl=identity_polars_expr,
        metadata={
            "category": "math",
            "summary": "Return the input unchanged.",
            "params": {
                "x": "Input value (scalar/array-like/expression).",
            },
            "return": "The same value as x.",
            "backend_support": {
                "numpy": "native",
                "polars": "native_expr",
            },
            "examples": [
                "jopera call math.identity --kwargs '{\"x\":3.14}'",
            ],
            "since": "1.2.0",
            "tags": ["math", "identity", "utility"],
        },
    ),
)



from __future__ import annotations

from ....core.spec import OperaFunction
from ..defs.core import chi2_cov_numpy, chi2_cov_polars_expr


STAT_CORE_DECLARATIONS: tuple[OperaFunction, ...] = (
    OperaFunction(
        namespace="stat",
        name="chi2_cov",
        arity=2,
        return_dtype=None,
        numpy_impl=chi2_cov_numpy,
        polars_expr_impl=chi2_cov_polars_expr,
        metadata={
            "category": "statistics",
            "summary": "Compute chi-square using residual vector(s) and covariance matrix.",
            "params": {
                "residual": "Residual vector or matrix (samples x observables).",
                "cov": "Covariance matrix.",
            },
            "return": "Chi-square scalar (single sample) or vector (multiple samples).",
            "backend_support": {
                "numpy": "native",
                "polars": "native_expr",
            },
            "examples": [
                "jopera call stat.chi2_cov --kwargs '{\"residual\":[1.0,0.0],\"cov\":[[2.0,0.0],[0.0,1.0]]}'",
            ],
            "since": "1.2.0",
            "tags": ["statistics", "chi2", "covariance"],
        },
    ),
)



from __future__ import annotations

import json
import textwrap
from pathlib import Path

import pytest

from jarvis_operas.errors import OperatorNotFound
from jarvis_operas.persistence import (
    delete_persisted_function,
    delete_persisted_namespace,
    get_overrides_store_path,
    get_sources_store_path,
    list_persisted_user_ops,
    persist_user_ops,
    update_persisted_function,
    update_persisted_namespace,
)


def test_global_registry_autoloads_persisted_user_ops(tmp_path, monkeypatch) -> None:
    op_file = tmp_path / "auto_ops.py"
    op_file.write_text(
        textwrap.dedent(
            """
            def square(x):
                return x * x

            __JARVIS_OPERAS__ = {
                "square": square,
            }
            """
        ),
        encoding="utf-8",
    )
    store_path = tmp_path / "persist_store.json"
    monkeypatch.setenv("JARVIS_OPERAS_PERSIST_FILE", str(store_path))

    persist_user_ops(str(op_file))
    entries = list_persisted_user_ops()
    assert len(entries) == 1
    assert entries[0]["path"] == str(op_file.resolve())
    assert entries[0]["namespace"] == "auto_ops"
    assert Path(entries[0]["load_path"]).exists()

    op_file.unlink()

    # Reset the singleton to emulate a fresh process bootstrap.
    import jarvis_operas.api as api

    monkeypatch.setattr(api, "_global_operas_registry", None)
    monkeypatch.setattr(api, "_global_operas", None)
    registry = api.get_global_operas_registry()

    assert registry.call("auto_ops.square", x=4) == 16


def test_persisted_function_delete_and_update_overrides(tmp_path, monkeypatch) -> None:
    op_file = tmp_path / "func_ops.py"
    op_file.write_text(
        textwrap.dedent(
            """
            def inc(x):
                return x + 1

            def dec(x):
                return x - 1

            __JARVIS_OPERAS__ = {
                "inc": inc,
                "dec": dec,
            }
            """
        ),
        encoding="utf-8",
    )
    store_path = tmp_path / "persist_store.json"
    monkeypatch.setenv("JARVIS_OPERAS_PERSIST_FILE", str(store_path))

    persist_user_ops(str(op_file))
    update_persisted_function("func_ops.inc", "renamed_ops.inc_user")
    delete_persisted_function("func_ops.dec")

    import jarvis_operas.api as api

    monkeypatch.setattr(api, "_global_operas_registry", None)
    monkeypatch.setattr(api, "_global_operas", None)
    registry = api.get_global_operas_registry()

    assert registry.call("renamed_ops.inc_user", x=2) == 3
    with pytest.raises(OperatorNotFound):
        registry.get("func_ops.inc")
    with pytest.raises(OperatorNotFound):
        registry.get("func_ops.dec")


def test_persisted_namespace_delete_and_update_overrides(tmp_path, monkeypatch) -> None:
    op_file = tmp_path / "ns_ops.py"
    op_file.write_text(
        textwrap.dedent(
            """
            def foo(x):
                return x

            __JARVIS_OPERAS__ = {
                "foo": foo,
            }
            """
        ),
        encoding="utf-8",
    )
    store_path = tmp_path / "persist_store.json"
    monkeypatch.setenv("JARVIS_OPERAS_PERSIST_FILE", str(store_path))

    persist_user_ops(str(op_file))
    update_persisted_namespace("ns_ops", "my_ns")

    import jarvis_operas.api as api

    monkeypatch.setattr(api, "_global_operas_registry", None)
    monkeypatch.setattr(api, "_global_operas", None)
    registry = api.get_global_operas_registry()
    assert registry.call("my_ns.foo", x=7) == 7

    delete_persisted_namespace("my_ns")
    monkeypatch.setattr(api, "_global_operas_registry", None)
    monkeypatch.setattr(api, "_global_operas", None)
    registry = api.get_global_operas_registry()

    with pytest.raises(OperatorNotFound):
        registry.get("my_ns.foo")


def test_persistence_rejects_protected_namespace_overrides(tmp_path, monkeypatch) -> None:
    store_path = tmp_path / "persist_store.json"
    monkeypatch.setenv("JARVIS_OPERAS_PERSIST_FILE", str(store_path))

    with pytest.raises(ValueError):
        delete_persisted_namespace("math")
    with pytest.raises(ValueError):
        update_persisted_namespace("user_ns", "helper")
    with pytest.raises(ValueError):
        delete_persisted_function("stat.chi2_cov")
    with pytest.raises(ValueError):
        update_persisted_function("user_ns.f", "math.f")


def test_persistence_splits_sources_and_overrides_files(tmp_path, monkeypatch) -> None:
    op_file = tmp_path / "split_ops.py"
    op_file.write_text(
        textwrap.dedent(
            """
            def f(x):
                return x + 1

            __JARVIS_OPERAS__ = {
                "f": f,
            }
            """
        ),
        encoding="utf-8",
    )
    store_path = tmp_path / "persist_store.json"
    monkeypatch.setenv("JARVIS_OPERAS_PERSIST_FILE", str(store_path))

    persist_user_ops(str(op_file))
    update_persisted_function("split_ops.f", "split_ns.f2")

    sources_path = get_sources_store_path()
    overrides_path = get_overrides_store_path()
    assert sources_path.exists()
    assert overrides_path.exists()
    assert sources_path.name == "sources.json"
    assert overrides_path.name == "overrides.json"

    sources_payload = json.loads(sources_path.read_text(encoding="utf-8"))
    overrides_payload = json.loads(overrides_path.read_text(encoding="utf-8"))
    assert isinstance(sources_payload.get("entries"), list)
    assert sources_payload["entries"][0]["path"] == str(op_file.resolve())
    assert "overrides" not in sources_payload
    assert overrides_payload["overrides"]["renamed_functions"]["split_ops.f"] == "split_ns.f2"
    assert "entries" not in overrides_payload


def test_persistence_can_read_legacy_combined_store(tmp_path, monkeypatch) -> None:
    op_file = tmp_path / "legacy_ops.py"
    op_file.write_text(
        textwrap.dedent(
            """
            def plus(x):
                return x + 5

            __JARVIS_OPERAS__ = {
                "plus": plus,
            }
            """
        ),
        encoding="utf-8",
    )
    legacy_store = tmp_path / "persist_store.json"
    legacy_store.write_text(
        json.dumps(
            {
                "version": 2,
                "entries": [{"path": str(op_file.resolve())}],
                "overrides": {
                    "deleted_functions": [],
                    "renamed_functions": {"legacy_ops.plus": "legacy_ns.plus2"},
                    "deleted_namespaces": [],
                    "renamed_namespaces": {},
                },
            },
            ensure_ascii=False,
            indent=2,
        ),
        encoding="utf-8",
    )
    monkeypatch.setenv("JARVIS_OPERAS_PERSIST_FILE", str(legacy_store))

    import jarvis_operas.api as api

    monkeypatch.setattr(api, "_global_operas_registry", None)
    monkeypatch.setattr(api, "_global_operas", None)
    registry = api.get_global_operas_registry()

    assert registry.call("legacy_ns.plus2", x=10) == 15
    with pytest.raises(OperatorNotFound):
        registry.get("legacy_ops.plus")



from __future__ import annotations

import importlib


def test_package_reload_does_not_bootstrap_globals(
    tmp_path,
    monkeypatch,
) -> None:
    monkeypatch.setenv("JARVIS_OPERAS_CURVE_INDEX", str(tmp_path / "curve-index.json"))
    monkeypatch.setenv("JARVIS_OPERAS_PERSIST_FILE", str(tmp_path / "persist-store.json"))

    import jarvis_operas as jo
    import jarvis_operas.api as api_mod

    monkeypatch.setattr(api_mod, "_global_operas_registry", None)
    monkeypatch.setattr(api_mod, "_global_operas", None)
    jo.__dict__.pop("operas_registry", None)
    jo.__dict__.pop("operas", None)

    importlib.reload(jo)

    assert api_mod._global_operas_registry is None
    assert api_mod._global_operas is None


def test_lazy_operas_exports_bootstrap_on_access(
    tmp_path,
    monkeypatch,
) -> None:
    monkeypatch.setenv("JARVIS_OPERAS_CURVE_INDEX", str(tmp_path / "curve-index.json"))
    monkeypatch.setenv("JARVIS_OPERAS_PERSIST_FILE", str(tmp_path / "persist-store.json"))

    import jarvis_operas as jo
    import jarvis_operas.api as api_mod

    monkeypatch.setattr(api_mod, "_global_operas_registry", None)
    monkeypatch.setattr(api_mod, "_global_operas", None)
    jo.__dict__.pop("operas_registry", None)
    jo.__dict__.pop("operas", None)
    importlib.reload(jo)

    assert api_mod._global_operas_registry is None
    assert api_mod._global_operas is None

    registry = jo.operas_registry
    assert registry is api_mod.get_global_operas_registry()
    assert api_mod._global_operas_registry is registry

    operas = jo.operas
    assert operas is api_mod.get_global_operas()
    assert api_mod._global_operas is operas



from __future__ import annotations

import re
from pathlib import Path
from typing import Any

FULL_NAME_SEPARATOR = "."
_NAMESPACE_FROM_STEM_RE = re.compile(r"[^0-9A-Za-z_]+")


def normalize_namespace(namespace: Any) -> str:
    normalized = str(namespace).strip()
    if not normalized:
        raise ValueError("namespace cannot be empty")
    if FULL_NAME_SEPARATOR in normalized:
        raise ValueError("namespace cannot contain '.'")
    return normalized


def normalize_short_name(name: Any, *, field_label: str = "name") -> str:
    normalized = str(name).strip()
    if not normalized:
        raise ValueError(f"{field_label} cannot be empty")
    if FULL_NAME_SEPARATOR in normalized:
        raise ValueError(f"{field_label} cannot contain '.'")
    return normalized


def split_full_name(value: Any) -> tuple[str, str]:
    normalized = str(value).strip()
    if not normalized:
        raise ValueError("full_name cannot be empty")
    if FULL_NAME_SEPARATOR not in normalized:
        raise ValueError("full_name must be in '<namespace>.<name>' format")
    namespace, short_name = normalized.split(FULL_NAME_SEPARATOR, 1)
    return normalize_namespace(namespace), short_name.strip()


def try_split_full_name(value: Any) -> tuple[str, str] | None:
    try:
        return split_full_name(value)
    except Exception:
        return None


def normalize_full_name(value: Any) -> str:
    namespace, short_name = split_full_name(value)
    normalized_short = normalize_short_name(short_name, field_label="operator name")
    return f"{namespace}{FULL_NAME_SEPARATOR}{normalized_short}"


def namespace_from_source_path(path: str | Path, *, default: str = "user_ops") -> str:
    stem = Path(path).stem.strip()
    normalized = _NAMESPACE_FROM_STEM_RE.sub("_", stem).strip("_")
    return normalized or default



from __future__ import annotations

from pathlib import Path


def test_maintaining_uses_current_module_layout() -> None:
    repo_root = Path(__file__).resolve().parent.parent
    body = (repo_root / "MAINTAINING.md").read_text(encoding="utf-8")

    assert "jarvis_operas/core/spec.py" in body
    assert "jarvis_operas/core/registry.py" in body
    assert "jarvis_operas/namespaces/*/decls/*.py" in body
    assert "scripts/generate_function_docs.py" in body

    assert "jarvis_operas/builtins/core.py" not in body
    assert "jarvis_operas/builtins/__init__.py" not in body
    assert "jarvis_operas/registry.py" not in body


def test_architecture_and_migration_docs_exist() -> None:
    repo_root = Path(__file__).resolve().parent.parent
    architecture = repo_root / "docs" / "architecture.md"
    migration = repo_root / "docs" / "migration.md"

    assert architecture.is_file()
    assert migration.is_file()

    architecture_body = architecture.read_text(encoding="utf-8")
    migration_body = migration.read_text(encoding="utf-8")

    assert "Stable Public Surface" in architecture_body
    assert "Mutation Policy" in architecture_body
    assert "Downstream Integration Guidance (HEP/PLOT)" in architecture_body

    assert "Import Path Updates" in migration_body
    assert "Downstream (Jarvis-HEP / Jarvis-PLOT) Adapter Checklist" in migration_body



from __future__ import annotations

import hashlib
import re
from types import SimpleNamespace
from typing import Any, Callable

import sympy as sp

from .api import get_global_operas_registry, is_global_operas_registry
from .core.registry import OperasRegistry
from .name_utils import try_split_full_name


def build_register_dicts(registry: OperasRegistry | None = None) -> dict[str, Callable[..., Any]]:
    """Return full-name -> numeric callable table for all registered functions."""

    target_registry = registry or get_global_operas_registry()
    table: dict[str, Callable[..., Any]] = {}
    for full_name in target_registry.list():
        try:
            declaration = target_registry.get(full_name)
        except Exception:
            continue

        fn = declaration.numpy_impl
        if callable(fn):
            table[full_name] = fn
    return table


def _to_symbolic_name(full_name: str) -> str:
    split = try_split_full_name(full_name)
    if split is None:
        raise ValueError(f"invalid full operator name: {full_name!r}")
    namespace, short_name = split
    safe_ns = re.sub(r"[^0-9A-Za-z_]+", "_", namespace).strip("_") or "ns"
    safe_short = re.sub(r"[^0-9A-Za-z_]+", "_", short_name).strip("_") or "fn"
    digest = hashlib.blake2b(full_name.encode("utf-8"), digest_size=4).hexdigest()
    symbolic = f"{safe_ns}__{safe_short}__{digest}"
    if symbolic[0].isdigit():
        symbolic = f"f_{symbolic}"
    return symbolic


def build_sympy_dicts(
    mapping: dict[str, Any] | None = None,
    *,
    namespaces: list[str] | None = None,
    include_all: bool = False,
) -> tuple[dict[str, Any], dict[str, Callable[..., Any]]]:
    """Build `sympify` locals and `lambdify` numeric maps from register dicts."""

    source = mapping if mapping is not None else build_register_dicts()
    parse_locals: dict[str, Any] = {}
    numeric_funcs_map: dict[str, Callable[..., Any]] = {}
    namespace_attrs: dict[str, dict[str, Any]] = {}
    allowed = set(namespaces or [])

    for full_name, fn in source.items():
        if not callable(fn):
            continue
        split = try_split_full_name(full_name)
        if split is None:
            continue
        namespace, short_name = split
        if allowed and not include_all and namespace not in allowed:
            continue
        symbolic_name = _to_symbolic_name(full_name)
        symbolic_fn = sp.Function(symbolic_name)
        namespace_attrs.setdefault(namespace, {})[short_name] = symbolic_fn
        numeric_funcs_map[symbolic_name] = fn

    for namespace, attrs in namespace_attrs.items():
        parse_locals[namespace] = SimpleNamespace(**attrs)

    return parse_locals, numeric_funcs_map


def _refresh_sympy_dicts() -> None:
    refreshed_func_locals, refreshed_numeric_funcs = build_sympy_dicts(
        build_register_dicts(),
        include_all=True,
    )
    func_locals.clear()
    func_locals.update(refreshed_func_locals)
    numeric_funcs.clear()
    numeric_funcs.update(refreshed_numeric_funcs)


def refresh_sympy_dicts_if_global_registry(registry: Any) -> bool:
    """Refresh SymPy dict snapshots only when mutating the global registry."""

    if not is_global_operas_registry(registry):
        return False
    _refresh_sympy_dicts()
    return True


func_locals, numeric_funcs = build_sympy_dicts(include_all=True)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from .api import get_global_operas, get_global_operas_registry, oper
from .curves import (
    add_interpolation_namespace,
    build_interpolation_declarations,
    get_curve_index_path,
    init_curve_cache,
    interpolation_manifest_resource,
    list_interpolation_namespaces,
    list_interpolation_namespace_entries,
    load_hot_curve_function_table,
    load_hot_curve_namespace_function_table,
    load_interpolation_manifest_library,
    remove_interpolation_namespace,
    register_hot_curves,
    register_hot_curve_namespace_in_registry,
    register_hot_curves_in_registry,
    validate_interpolation_namespace_index,
)
from .core import (
    OperaFunction,
    Operas,
    OperasRegistry,
)
from .errors import (
    OperatorCallError,
    OperatorConflict,
    OperatorLoadError,
    OperatorNotFound,
)
from .loading import discover_entrypoints, load_user_ops
from .logging import configure_cli_logger, get_log_mode, get_logger, set_log_mode
from .persistence import (
    apply_persisted_overrides,
    delete_persisted_function,
    delete_persisted_namespace,
    get_overrides_store_path,
    get_persist_store_path,
    get_sources_store_path,
    list_persisted_user_ops,
    persist_user_ops,
    update_persisted_function,
    update_persisted_namespace,
)

if TYPE_CHECKING:
    from .integration import build_register_dicts, build_sympy_dicts, func_locals, numeric_funcs

    operas_registry: OperasRegistry
    operas: Operas

_LAZY_INTEGRATION_EXPORTS = frozenset(
    {
        "build_register_dicts",
        "build_sympy_dicts",
        "func_locals",
        "numeric_funcs",
    }
)

__all__ = [
    "OperaFunction",
    "OperasRegistry",
    "Operas",
    "OperatorNotFound",
    "OperatorConflict",
    "OperatorLoadError",
    "OperatorCallError",
    "build_register_dicts",
    "build_sympy_dicts",
    "get_logger",
    "set_log_mode",
    "get_log_mode",
    "configure_cli_logger",
    "oper",
    "operas_registry",
    "get_global_operas_registry",
    "operas",
    "get_global_operas",
    "load_user_ops",
    "discover_entrypoints",
    "persist_user_ops",
    "list_persisted_user_ops",
    "get_persist_store_path",
    "get_sources_store_path",
    "get_overrides_store_path",
    "apply_persisted_overrides",
    "delete_persisted_function",
    "update_persisted_function",
    "delete_persisted_namespace",
    "update_persisted_namespace",
    "get_curve_index_path",
    "build_interpolation_declarations",
    "add_interpolation_namespace",
    "init_curve_cache",
    "interpolation_manifest_resource",
    "list_interpolation_namespaces",
    "list_interpolation_namespace_entries",
    "load_hot_curve_function_table",
    "load_hot_curve_namespace_function_table",
    "load_interpolation_manifest_library",
    "remove_interpolation_namespace",
    "func_locals",
    "numeric_funcs",
    "register_hot_curves",
    "register_hot_curve_namespace_in_registry",
    "register_hot_curves_in_registry",
    "validate_interpolation_namespace_index",
]


def __getattr__(name: str) -> Any:
    if name == "operas_registry":
        return get_global_operas_registry()
    if name == "operas":
        return get_global_operas()
    if name in _LAZY_INTEGRATION_EXPORTS:
        from . import integration as _integration

        return getattr(_integration, name)
    raise AttributeError(f"module '{__name__}' has no attribute '{name}'")


def __dir__() -> list[str]:
    return sorted(set(globals()) | _LAZY_INTEGRATION_EXPORTS | {"operas_registry", "operas"})



from __future__ import annotations

import argparse
import asyncio
import difflib
import json
import sys
import textwrap
from importlib.metadata import PackageNotFoundError, version as dist_version
from typing import Any

from .api import get_global_operas_registry
from .core.registry import OperasRegistry
from .core.spec import OperaFunction
from .curves import (
    add_interpolation_namespace,
    init_builtin_curve_cache,
    init_curve_cache,
    list_interpolation_namespace_entries,
    register_hot_curves_in_registry,
    remove_interpolation_namespace,
    validate_interpolation_namespace_index,
)
from .errors import OperatorNotFound
from .integration import refresh_sympy_dicts_if_global_registry
from .loading import load_user_ops
from .logging import configure_cli_logger
from .mutation_policy import ensure_cli_write_allowed
from .name_utils import try_split_full_name
from .namespace_policy import ensure_user_namespace_allowed, is_protected_namespace
from .persistence import (
    apply_persisted_overrides,
    clear_persisted_function_overrides,
    clear_persisted_namespace_overrides,
    delete_persisted_function,
    delete_persisted_namespace,
    get_overrides_store_path,
    get_sources_store_path,
    persist_user_ops,
)

_LOG_MODE_CHOICES = ("warning", "info", "debug")


def _resolve_version() -> str:
    try:
        return dist_version("Jarvis-Operas")
    except PackageNotFoundError:
        return "dev"


class _JarvisHelpFormatter(argparse.RawDescriptionHelpFormatter):
    """Hide argparse subparser metavar line and keep only concrete commands."""

    def _format_action(self, action: argparse.Action) -> str:
        if isinstance(action, argparse._SubParsersAction):
            parts: list[str] = []
            self._indent()
            for subaction in self._iter_indented_subactions(action):
                parts.append(self._format_action(subaction))
            self._dedent()
            return "".join(parts)
        return super()._format_action(action)


def _root_card() -> str:
    return textwrap.dedent(
        """
        Jarvis-Operas CLI

        Start here:
          jopera init
          jopera list
          jopera info helper.eggbox
          jopera call math.add --kwargs '{"a":1,"b":2}'
          jopera init --manifest ./manifest.json
          jopera load /path/to/my_ops.py

        Namespaces:
          math.*    numeric helpers
          stat.*    statistics functions
          helper.*  HEP scan helpers

        Use 'jopera --help' for full options, 'jopera help examples' for copy-paste examples.
        """
    ).strip()


def _examples_card() -> str:
    return textwrap.dedent(
        """
        Quick examples:
          jopera init
          jopera init --manifest ./manifest.json
          jopera list
          jopera list --namespace helper
          jopera info stat.chi2_cov
          jopera call math.add --kwargs '{"a":1,"b":2}'
          jopera call helper.eggbox --kwargs '{"x":0.5,"y":0.0}'
          jopera load ./my_ops.py
          jopera call my_ops.my_func --arg x=1 --arg y=2
          jopera interp list
          jopera interp validate
        """
    ).strip()


def _advanced_card() -> str:
    return textwrap.dedent(
        """
        Advanced options:
          -d, --debug         Enable debug logging (same as --log-mode debug)
          --user-ops PATH     Load operators from Python file (repeatable)
          --session-only      (load command) do not persist this source path
          --session-only      (delete/update commands) do not persist override rules
          --arg key=value     Append call kwargs (JSON-decoded when possible)
          --kwargs JSON       Provide call kwargs as JSON object
          --backend BACKEND   call/acall backend: numpy|polars (default: numpy)
          --log-mode MODE     warning|info|debug
          --json              Machine-readable output for supported commands
          init --force        Rebuild all curve cache pickle artifacts
          init --namespaces   Comma list of interpolation namespaces (e.g. dmdd,interp1)
          write policy        User write path is only 'jopera load'
          developer mode      Set JARVIS_OPERAS_DEV_WRITE=1 to enable developer write commands

        Example:
          jopera call my_ops.my_func --user-ops ./my_ops.py --arg x=1 --arg y=2 --log-mode debug
        """
    ).strip()


def _parse_kv_arg(raw: str) -> tuple[str, Any]:
    if "=" not in raw:
        raise argparse.ArgumentTypeError(
            f"Invalid --arg '{raw}'. Use key=value format."
        )
    key, raw_value = raw.split("=", 1)
    key = key.strip()
    if not key:
        raise argparse.ArgumentTypeError("Argument key cannot be empty.")

    try:
        value = json.loads(raw_value)
    except json.JSONDecodeError:
        value = raw_value

    return key, value


def _parse_csv_namespaces(raw: str | None) -> list[str] | None:
    if raw is None:
        return None
    names = [item.strip() for item in raw.split(",") if item.strip()]
    return names or None


def _merge_kwargs(raw_json: str | None, key_values: list[tuple[str, Any]]) -> dict[str, Any]:
    kwargs: dict[str, Any] = {}
    if raw_json:
        parsed = json.loads(raw_json)
        if not isinstance(parsed, dict):
            raise ValueError("--kwargs must decode to a JSON object")
        kwargs.update(parsed)

    for key, value in key_values:
        kwargs[key] = value

    return kwargs


def _print_value(value: Any, as_json: bool) -> None:
    if as_json:
        print(json.dumps(value, ensure_ascii=False, indent=2, default=str))
        return

    if isinstance(value, (dict, list, tuple)):
        print(json.dumps(value, ensure_ascii=False, indent=2, default=str))
        return

    print(value)


def _build_list_entries(registry, names: list[str]) -> list[dict[str, Any]]:
    entries: list[dict[str, Any]] = []
    for full_name in names:
        info = registry.info(full_name)
        entries.append(
            {
                "id": info["id"],
                "name": info["name"],
                "namespace": info["namespace"],
            }
        )
    return sorted(
        entries,
        key=lambda item: (
            str(item.get("namespace", "")).lower(),
            str(item.get("name", "")).lower(),
        ),
    )


def _group_entries_by_namespace(
    entries: list[dict[str, Any]],
) -> list[tuple[str, list[dict[str, Any]]]]:
    grouped: dict[str, list[dict[str, Any]]] = {}
    for entry in entries:
        namespace = entry["namespace"]
        if namespace not in grouped:
            grouped[namespace] = []
        grouped[namespace].append(entry)
    ordered_namespaces = sorted(grouped.keys(), key=lambda item: item.lower())
    return [(namespace, grouped[namespace]) for namespace in ordered_namespaces]


def _print_list_human(entries: list[dict[str, Any]], namespace_filter: str | None) -> None:
    if not entries:
        if namespace_filter:
            print(f"No functions found in namespace '{namespace_filter}'.")
        else:
            print("No functions found.")
        return

    lines: list[str] = []
    groups = _group_entries_by_namespace(entries)
    for index, (namespace, items) in enumerate(groups):
        lines.append(f"{namespace} ({len(items)})")
        for item in items:
            lines.append(f"  - {item['id']}\t{item['name']}")
        if index != len(groups) - 1:
            lines.append("")

    print("\n".join(lines))


def _print_interp_list_human(
    entries: list[dict[str, Any]],
    manifest_path: str,
) -> None:
    if not entries:
        print(f"No interpolation namespaces found in: {manifest_path}")
        return

    lines = [
        f"Interpolation namespaces ({len(entries)})",
        f"Manifest:\t{manifest_path}",
        "",
    ]
    for entry in entries:
        line = f"- {entry['namespace']}\t{entry['manifest']}"
        description = entry.get("description")
        if isinstance(description, str) and description.strip():
            line = f"{line}\t# {description.strip()}"
        lines.append(line)
    print("\n".join(lines))


def _extract_cli_example(metadata: Any) -> str | None:
    if not isinstance(metadata, dict):
        return None
    examples = metadata.get("examples")
    if isinstance(examples, list):
        for item in examples:
            if isinstance(item, str) and item.strip():
                return item.strip()
            if isinstance(item, dict):
                cli = item.get("cli")
                if isinstance(cli, str) and cli.strip():
                    return cli.strip()
    if isinstance(examples, dict):
        cli = examples.get("cli")
        if isinstance(cli, str) and cli.strip():
            return cli.strip()
    return None


def _is_user_loaded_function(info: dict[str, Any]) -> bool:
    namespace = info.get("namespace")
    if not isinstance(namespace, str) or not namespace.strip():
        return False
    try:
        return not is_protected_namespace(namespace)
    except Exception:
        return False


def _print_info_human(info: dict[str, Any]) -> None:
    metadata = info.get("metadata", {})
    supports_async = bool(info.get("supports_async", info.get("is_async")))
    note = metadata.get("note") if isinstance(metadata, dict) else None
    lines = [
        f"Name:\t{info.get('name', '')}",
        f"ID:\t{info.get('id', '')}",
        f"Namespace:\t{info.get('namespace', '')}",
        f"Signature:\t{info.get('signature', '')}",
        f"Async:\t{str(supports_async).lower()}",
    ]
    category = metadata.get("category")
    if category:
        lines.append(f"Category:\t{category}")
    doc = info.get("docstring", "")
    if doc:
        lines.append(f"Summary:\t{doc}")
    if isinstance(note, str) and note.strip():
        lines.append(f"Note:\t{note.strip()}")
    if _is_user_loaded_function(info):
        lines.append(f"UserSources:\t{get_sources_store_path()}")
        lines.append(f"UserOverrides:\t{get_overrides_store_path()}")
    suggested = _extract_cli_example(metadata)
    if isinstance(suggested, str) and suggested.strip():
        lines.append("")
        lines.append("Try next:")
        lines.append(f"\t{suggested.strip()}")
    print("\n".join(lines))


def _print_not_found_error(full_name: str, names: list[str]) -> None:
    suggestions = difflib.get_close_matches(full_name, names, n=3, cutoff=0.25)
    print(f"Function '{full_name}' not found.", file=sys.stderr)
    if suggestions:
        print("Did you mean:", file=sys.stderr)
        for item in suggestions:
            print(f"  - {item}", file=sys.stderr)
    split = try_split_full_name(full_name)
    if split is not None:
        namespace, _ = split
        print("Try:", file=sys.stderr)
        print(f"  jopera list --namespace {namespace}", file=sys.stderr)
    else:
        print("Try:", file=sys.stderr)
        print("  jopera list", file=sys.stderr)


def _load_sources(user_ops: list[str]) -> list[str]:
    registry = get_global_operas_registry()
    loaded: list[str] = []

    for path in user_ops:
        loaded.extend(load_user_ops(path, registry, refresh_sympy=False))

    if user_ops:
        apply_persisted_overrides(registry)
        refresh_sympy_dicts_if_global_registry(registry)

    return sorted(set(loaded))


def _registry_with_user_sources(user_ops: list[str]) -> OperasRegistry:
    _load_sources(user_ops)
    return get_global_operas_registry()


def _ensure_optional_user_namespace(namespace: str | None, *, action: str) -> None:
    if namespace is None:
        return
    ensure_user_namespace_allowed(
        namespace,
        action=action,
    )


def _refresh_global_registry_views(registry: OperasRegistry) -> None:
    refresh_sympy_dicts_if_global_registry(registry)


def _cmd_list(args: argparse.Namespace) -> int:
    registry = _registry_with_user_sources(args.user_ops)
    names = registry.list(namespace=args.namespace)
    entries = _build_list_entries(registry, names)
    if args.json:
        _print_value(entries, as_json=True)
    else:
        _print_list_human(entries, args.namespace)
    return 0


def _cmd_info(args: argparse.Namespace) -> int:
    registry = _registry_with_user_sources(args.user_ops)
    info = registry.info(args.target)
    if args.json and _is_user_loaded_function(info):
        info = dict(info)
        info["user_store"] = {
            "sources_path": str(get_sources_store_path()),
            "overrides_path": str(get_overrides_store_path()),
        }
    if args.json:
        _print_value(info, as_json=True)
    else:
        _print_info_human(info)
    return 0


def _cmd_call_common(args: argparse.Namespace, *, async_mode: bool) -> int:
    registry = _registry_with_user_sources(args.user_ops)
    kwargs = _merge_kwargs(args.kwargs, args.arg)

    if async_mode:
        result = asyncio.run(registry.acall(args.full_name, backend=args.backend, **kwargs))
    else:
        result = registry.call(args.full_name, backend=args.backend, **kwargs)
    _print_value(result, as_json=args.json)
    return 0


def _cmd_call(args: argparse.Namespace) -> int:
    return _cmd_call_common(args, async_mode=False)


def _cmd_acall(args: argparse.Namespace) -> int:
    return _cmd_call_common(args, async_mode=True)


def _cmd_load(args: argparse.Namespace) -> int:
    registry = get_global_operas_registry()
    _ensure_optional_user_namespace(
        args.namespace,
        action="CLI load namespace override",
    )
    loaded = load_user_ops(
        args.path,
        registry,
        namespace=args.namespace,
        refresh_sympy=False,
    )
    if not args.session_only:
        persist_user_ops(args.path, namespace=args.namespace)
    _refresh_global_registry_views(registry)
    _print_value(loaded, as_json=True)
    return 0


def _register_from_registry(
    source_registry: OperasRegistry,
    target_registry: OperasRegistry,
    full_name: str,
) -> None:
    source_decl = source_registry.get(full_name)
    target_registry.register(
        OperaFunction(
            namespace=source_decl.namespace,
            name=source_decl.name,
            arity=source_decl.arity,
            return_dtype=source_decl.return_dtype,
            numpy_impl=source_decl.numpy_impl,
            polars_expr_impl=source_decl.polars_expr_impl,
            flags=source_decl.flags,
            metadata=dict(source_decl.metadata),
        )
    )


def _cmd_update(args: argparse.Namespace) -> int:
    ensure_cli_write_allowed("update")
    registry = get_global_operas_registry()
    _ensure_optional_user_namespace(
        args.namespace,
        action="CLI update namespace override",
    )
    staging = OperasRegistry()
    loaded = load_user_ops(
        args.path,
        staging,
        namespace=args.namespace,
        refresh_sympy=False,
    )

    if args.function:
        targets: list[str] = []
        for full_name in loaded:
            split = try_split_full_name(full_name)
            if split is None:
                continue
            namespace, short_name = split
            if short_name != args.function:
                continue
            if args.namespace is not None and namespace != args.namespace:
                continue
            targets.append(full_name)
        if not targets:
            raise ValueError(
                f"Function '{args.function}' was not found in script '{args.path}'."
            )
        if len(targets) > 1:
            raise ValueError(
                f"Function '{args.function}' appears in multiple namespaces. "
                "Use --namespace to disambiguate."
            )
    else:
        targets = list(loaded)

    if not args.function:
        target_namespaces = sorted(
            {
                split[0]
                for name in targets
                for split in [try_split_full_name(name)]
                if split is not None
            }
        )
        for namespace in target_namespaces:
            registry.delete_namespace(namespace)
            if not args.session_only:
                clear_persisted_namespace_overrides(namespace)
    else:
        target_namespaces = []

    updated: list[str] = []
    for full_name in targets:
        if args.function:
            try:
                registry.delete(full_name)
            except OperatorNotFound:
                pass
            if not args.session_only:
                clear_persisted_function_overrides(full_name)
                split = try_split_full_name(full_name)
                if split is not None:
                    clear_persisted_namespace_overrides(split[0])

        _register_from_registry(staging, registry, full_name)
        updated.append(full_name)

    if not args.session_only:
        persist_user_ops(args.path, namespace=args.namespace)

    _refresh_global_registry_views(registry)
    _print_value(
        {
            "updated": updated,
            "mode": "function" if args.function else "namespace",
            "session_only": bool(args.session_only),
        },
        as_json=True,
    )
    return 0


def _cmd_help(args: argparse.Namespace) -> int:
    topic = args.topic or "examples"
    if topic == "examples":
        print(_examples_card())
        return 0
    if topic == "advanced":
        print(_advanced_card())
        return 0

    print(f"Unknown help topic: {topic}", file=sys.stderr)
    return 1


def _cmd_init(args: argparse.Namespace) -> int:
    selected_namespaces = _parse_csv_namespaces(args.namespaces)
    if args.manifest:
        summary = init_curve_cache(
            args.manifest,
            source_root=args.source_root,
            cache_root=args.cache_root,
            index_path=args.index_path,
            namespaces=selected_namespaces,
            force=bool(args.force),
        )
    else:
        summary = init_builtin_curve_cache(
            source_root=args.source_root,
            cache_root=args.cache_root,
            index_path=args.index_path,
            namespaces=selected_namespaces,
            force=bool(args.force),
        )
    register_hot_curves_in_registry(
        get_global_operas_registry(),
        index_path=summary["index_path"],
        namespaces=selected_namespaces,
    )
    if args.json:
        _print_value(summary, as_json=True)
        return 0

    print(f"Manifest: {summary['manifest_path']}")
    print(f"Index:    {summary['index_path']}")
    print(f"Total:    {summary['total']}")
    print(f"Compiled: {len(summary['compiled'])}")
    print(f"Cached:   {len(summary['cached'])}")
    return 0


def _cmd_interp_list(args: argparse.Namespace) -> int:
    entries = list_interpolation_namespace_entries(manifest_path=args.manifest)
    display_manifest = args.manifest or "<built-in>"
    if args.json:
        _print_value(
            {
                "manifest_path": display_manifest,
                "namespace_count": len(entries),
                "namespaces": entries,
            },
            as_json=True,
        )
        return 0

    _print_interp_list_human(entries, manifest_path=display_manifest)
    return 0


def _cmd_interp_validate(args: argparse.Namespace) -> int:
    report = validate_interpolation_namespace_index(manifest_path=args.manifest)
    if args.json:
        _print_value(report, as_json=True)
    else:
        status = "ok" if report.get("ok") else "failed"
        print(f"Validation: {status}")
        print(f"Manifest:   {report.get('manifest_path')}")
        print(f"Namespaces: {report.get('namespace_count')}")
        print(f"Functions:  {report.get('function_count')}")
        errors = report.get("errors", [])
        if isinstance(errors, list) and errors:
            print("")
            print("Errors:")
            for item in errors:
                print(f"- {item}")
    return 0 if bool(report.get("ok")) else 1


def _cmd_interp_add(args: argparse.Namespace) -> int:
    ensure_cli_write_allowed("interp add")
    result = add_interpolation_namespace(
        manifest_path=args.manifest,
        namespace=args.namespace,
        namespace_manifest=args.namespace_manifest,
        description=args.description,
        overwrite=bool(args.overwrite),
    )
    if args.json:
        _print_value(result, as_json=True)
        return 0

    state = "added"
    if result.get("updated"):
        state = "updated"
    if result.get("unchanged"):
        state = "unchanged"
    print(f"Namespace {state}: {result['namespace']}")
    print(f"Manifest: {result['manifest_path']}")
    return 0


def _cmd_interp_remove(args: argparse.Namespace) -> int:
    ensure_cli_write_allowed("interp remove")
    result = remove_interpolation_namespace(
        manifest_path=args.manifest,
        namespace=args.namespace,
    )
    if args.json:
        _print_value(result, as_json=True)
    else:
        if result.get("removed"):
            print(f"Namespace removed: {result['namespace']}")
        else:
            print(f"Namespace not found: {result['namespace']}")
        print(f"Manifest: {result['manifest_path']}")

    if args.strict and not bool(result.get("removed")):
        return 1
    return 0


def _cmd_delete_func(args: argparse.Namespace) -> int:
    ensure_cli_write_allowed("delete-func")
    registry = get_global_operas_registry()
    resolved_full_name = registry.resolve_name(args.target)
    split = try_split_full_name(resolved_full_name)
    if split is not None and is_protected_namespace(split[0]):
        raise ValueError(
            f"namespace '{split[0]}' is protected and cannot be modified by delete-func"
        )
    registry.delete(args.target)
    if not args.session_only:
        delete_persisted_function(resolved_full_name)
    _refresh_global_registry_views(registry)
    _print_value({"deleted": resolved_full_name, "session_only": bool(args.session_only)}, as_json=True)
    return 0


def _cmd_delete_namespace(args: argparse.Namespace) -> int:
    ensure_cli_write_allowed("delete-namespace")
    ensure_user_namespace_allowed(
        args.namespace,
        action="CLI delete-namespace",
    )
    registry = get_global_operas_registry()
    deleted = registry.delete_namespace(args.namespace)
    if not args.session_only:
        delete_persisted_namespace(args.namespace)
    _refresh_global_registry_views(registry)
    _print_value(
        {
            "namespace": args.namespace,
            "deleted": deleted,
            "session_only": bool(args.session_only),
        },
        as_json=True,
    )
    return 0


def _add_source_args(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--user-ops",
        action="append",
        default=[],
        help="Path to a Python file with user operators. Can be provided multiple times.",
    )


def _add_log_mode_arg(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--log-mode",
        choices=_LOG_MODE_CHOICES,
        default="warning",
        help="Console log verbosity (default: warning).",
    )


def _add_call_args(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "full_name",
        help="Full function name, e.g. math.add or my_ops.my_op",
    )
    parser.add_argument(
        "--kwargs",
        default=None,
        help="JSON object for operator kwargs, e.g. '{\"a\": 1, \"b\": 2}'",
    )
    parser.add_argument(
        "--arg",
        action="append",
        type=_parse_kv_arg,
        default=[],
        help="Additional key=value argument. Value uses JSON decode when possible.",
    )
    parser.add_argument(
        "--json",
        action="store_true",
        help="Force JSON output formatting.",
    )
    parser.add_argument(
        "--backend",
        choices=("numpy", "polars"),
        default="numpy",
        help="Execution backend for this call (default: numpy).",
    )
    _add_source_args(parser)


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="jopera",
        description=textwrap.dedent(
            """
            Jarvis-Operas CLI
            Register, inspect, and run functions by name.
            """
        ).strip(),
        epilog="Use 'jopera help examples' for copy-paste examples.",
        formatter_class=_JarvisHelpFormatter,
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"%(prog)s {_resolve_version()}",
        help="Show jopera version and exit.",
    )
    parser.add_argument(
        "--help-advanced",
        action="store_true",
        help="Show advanced options and exit.",
    )
    parser.add_argument(
        "-d",
        "--debug",
        action="store_true",
        help="Enable debug logging (equivalent to --log-mode debug).",
    )
    subparsers = parser.add_subparsers(
        dest="command",
        required=False,
        title="commands",
    )

    parser_list = subparsers.add_parser("list", help="List available functions")
    parser_list.add_argument("--namespace", default=None, help="Filter by namespace")
    parser_list.add_argument(
        "--json",
        action="store_true",
        help="Output JSON instead of grouped text.",
    )
    _add_log_mode_arg(parser_list)
    _add_source_args(parser_list)
    parser_list.set_defaults(func=_cmd_list)

    parser_info = subparsers.add_parser("info", help="Show function metadata")
    parser_info.add_argument("target", help="Full function name or function id")
    parser_info.add_argument("--json", action="store_true", help="Force JSON output")
    _add_log_mode_arg(parser_info)
    _add_source_args(parser_info)
    parser_info.set_defaults(func=_cmd_info)

    parser_call = subparsers.add_parser("call", help="Call a function synchronously")
    _add_log_mode_arg(parser_call)
    _add_call_args(parser_call)
    parser_call.set_defaults(func=_cmd_call)

    parser_acall = subparsers.add_parser("acall", help="Call a function asynchronously")
    _add_log_mode_arg(parser_acall)
    _add_call_args(parser_acall)
    parser_acall.set_defaults(func=_cmd_acall)

    parser_load = subparsers.add_parser("load", help="Load user functions from a file")
    parser_load.add_argument("path", help="Path to Python file")
    _add_log_mode_arg(parser_load)
    parser_load.add_argument(
        "--namespace",
        default=None,
        help="Override namespace. Default derives from script filename.",
    )
    parser_load.add_argument(
        "--session-only",
        action="store_true",
        help="Load only for this process. Skip persistent registration.",
    )
    parser_load.set_defaults(func=_cmd_load)

    parser_init = subparsers.add_parser(
        "init",
        help="Compile curve manifest into local runtime cache",
    )
    _add_log_mode_arg(parser_init)
    parser_init.add_argument(
        "--manifest",
        default=None,
        help="Path to manifest JSON defining curves.",
    )
    parser_init.add_argument(
        "--namespaces",
        default=None,
        help="Optional comma-separated interpolation namespaces to process.",
    )
    parser_init.add_argument(
        "--source-root",
        default=None,
        help="Optional base directory for relative curve JSON paths.",
    )
    parser_init.add_argument(
        "--cache-root",
        default=None,
        help="Directory for generated pickle cache and index.json.",
    )
    parser_init.add_argument(
        "--index-path",
        default=None,
        help="Optional explicit index.json path (overrides --cache-root).",
    )
    parser_init.add_argument(
        "--force",
        action="store_true",
        help="Force recompilation even when hash matches existing cache.",
    )
    parser_init.add_argument(
        "--json",
        action="store_true",
        help="Output JSON summary.",
    )
    parser_init.set_defaults(func=_cmd_init)

    parser_interp = subparsers.add_parser(
        "interp",
        help="Manage interpolation namespace manifest index",
    )
    _add_log_mode_arg(parser_interp)
    interp_subparsers = parser_interp.add_subparsers(
        dest="interp_command",
        required=True,
        title="interp commands",
    )

    parser_interp_list = interp_subparsers.add_parser(
        "list",
        help="List interpolation namespaces from root manifest",
    )
    parser_interp_list.add_argument(
        "--manifest",
        default=None,
        help="Path to root interpolation manifest index (default: built-in).",
    )
    parser_interp_list.add_argument(
        "--json",
        action="store_true",
        help="Output JSON summary.",
    )
    parser_interp_list.set_defaults(func=_cmd_interp_list)

    parser_interp_validate = interp_subparsers.add_parser(
        "validate",
        help="Validate root interpolation manifest index and namespace manifests",
    )
    parser_interp_validate.add_argument(
        "--manifest",
        default=None,
        help="Path to root interpolation manifest index (default: built-in).",
    )
    parser_interp_validate.add_argument(
        "--json",
        action="store_true",
        help="Output JSON validation report.",
    )
    parser_interp_validate.set_defaults(func=_cmd_interp_validate)

    parser_interp_add = interp_subparsers.add_parser(
        "add",
        help="Add one interpolation namespace entry to root manifest index",
    )
    parser_interp_add.add_argument(
        "--manifest",
        required=True,
        help="Writable root interpolation manifest index path.",
    )
    parser_interp_add.add_argument(
        "namespace",
        help="Namespace to add, e.g. dmdd.",
    )
    parser_interp_add.add_argument(
        "namespace_manifest",
        help="Namespace manifest path or ref, relative to root manifest directory.",
    )
    parser_interp_add.add_argument(
        "--description",
        default=None,
        help="Optional namespace description.",
    )
    parser_interp_add.add_argument(
        "--overwrite",
        action="store_true",
        help="Replace existing namespace entry when namespace already exists.",
    )
    parser_interp_add.add_argument(
        "--json",
        action="store_true",
        help="Output JSON summary.",
    )
    parser_interp_add.set_defaults(func=_cmd_interp_add)

    parser_interp_remove = interp_subparsers.add_parser(
        "remove",
        help="Remove one interpolation namespace entry from root manifest index",
    )
    parser_interp_remove.add_argument(
        "--manifest",
        required=True,
        help="Writable root interpolation manifest index path.",
    )
    parser_interp_remove.add_argument(
        "namespace",
        help="Namespace to remove.",
    )
    parser_interp_remove.add_argument(
        "--strict",
        action="store_true",
        help="Return non-zero when namespace does not exist.",
    )
    parser_interp_remove.add_argument(
        "--json",
        action="store_true",
        help="Output JSON summary.",
    )
    parser_interp_remove.set_defaults(func=_cmd_interp_remove)

    parser_update = subparsers.add_parser(
        "update",
        help="Update functions from a Python file (default: whole namespace)",
    )
    parser_update.add_argument("path", help="Path to Python file")
    _add_log_mode_arg(parser_update)
    parser_update.add_argument(
        "--namespace",
        default=None,
        help="Override namespace. Default derives from script filename.",
    )
    parser_update.add_argument(
        "--function",
        default=None,
        help="Only update one function short name from this script.",
    )
    parser_update.add_argument(
        "--session-only",
        action="store_true",
        help="Update only for this process. Skip persistent override updates.",
    )
    parser_update.set_defaults(func=_cmd_update)

    parser_delete_func = subparsers.add_parser(
        "delete-func",
        help="Delete one function by full name or id",
    )
    _add_log_mode_arg(parser_delete_func)
    parser_delete_func.add_argument("target", help="Full function name or function id")
    parser_delete_func.add_argument(
        "--session-only",
        action="store_true",
        help="Delete only for this process. Skip persistent deletion rule.",
    )
    parser_delete_func.set_defaults(func=_cmd_delete_func)

    parser_delete_namespace = subparsers.add_parser(
        "delete-namespace",
        help="Delete all functions in a namespace",
    )
    _add_log_mode_arg(parser_delete_namespace)
    parser_delete_namespace.add_argument("namespace", help="Namespace to delete")
    parser_delete_namespace.add_argument(
        "--session-only",
        action="store_true",
        help="Delete only for this process. Skip persistent deletion rule.",
    )
    parser_delete_namespace.set_defaults(func=_cmd_delete_namespace)

    parser_help = subparsers.add_parser(
        "help",
        help="Show quick examples and tips",
    )
    parser_help.add_argument(
        "topic",
        nargs="?",
        default="examples",
        choices=("examples", "advanced"),
        help="Help topic",
    )
    parser_help.set_defaults(func=_cmd_help)

    return parser


def main(argv: list[str] | None = None) -> int:
    raw_argv = list(sys.argv[1:] if argv is None else argv)
    force_debug = False
    parse_argv: list[str] = []
    for token in raw_argv:
        if token in ("-d", "--debug"):
            force_debug = True
            continue
        parse_argv.append(token)

    parser = build_parser()
    args = parser.parse_args(parse_argv)

    if args.help_advanced:
        print(_advanced_card())
        return 0

    if args.command is None:
        print(_root_card())
        return 0

    selected_mode = "debug" if force_debug else getattr(args, "log_mode", "warning")
    configure_cli_logger(selected_mode)

    try:
        return args.func(args)
    except OperatorNotFound as exc:
        registry = get_global_operas_registry()
        _print_not_found_error(exc.full_name, registry.list())
        return 1
    except Exception as exc:
        print(f"jopera error: {exc}", file=sys.stderr)
        return 1


if __name__ == "__main__":
    raise SystemExit(main())



from __future__ import annotations

import textwrap

from jarvis_operas import OperaFunction, OperasRegistry, load_user_ops


def test_load_user_ops_with_decorator_into_operas_registry(tmp_path) -> None:
    op_file = tmp_path / "decor_ops.py"
    op_file.write_text(
        textwrap.dedent(
            """
            from jarvis_operas import oper

            @oper("double", arity=1, tag="demo")
            def double(x):
                return x * 2
            """
        ),
        encoding="utf-8",
    )

    registry = OperasRegistry()
    loaded = load_user_ops(str(op_file), registry)

    assert loaded == ["decor_ops.double"]
    assert registry.call("decor_ops.double", 6) == 12
    info = registry.info("decor_ops.double")
    assert info["arity"] == 1
    assert info["metadata"]["tag"] == "demo"


def test_load_user_ops_with_export_whitelist_into_operas_registry(tmp_path) -> None:
    op_file = tmp_path / "whitelist_ops.py"
    op_file.write_text(
        textwrap.dedent(
            """
            def triple(x):
                return x * 3

            __JARVIS_OPERAS__ = {
                "triple": triple,
            }
            """
        ),
        encoding="utf-8",
    )

    registry = OperasRegistry()
    loaded = load_user_ops(str(op_file), registry)

    assert loaded == ["whitelist_ops.triple"]
    assert registry.call("whitelist_ops.triple", 5) == 15
    info = registry.info("whitelist_ops.triple")
    assert info["metadata"]["source"] == "__JARVIS_OPERAS__"


def test_load_user_ops_accepts_operafunction_values(tmp_path) -> None:
    op_file = tmp_path / "ops_decl.py"
    op_file.write_text(
        textwrap.dedent(
            """
            from jarvis_operas import OperaFunction

            def scale_three(x):
                return x * 3

            __JARVIS_OPERAS__ = {
                "scale_three": OperaFunction(
                    namespace="placeholder",
                    name="placeholder",
                    arity=1,
                    return_dtype=None,
                    numpy_impl=scale_three,
                    metadata={"kind": "decl"},
                ),
            }
            """
        ),
        encoding="utf-8",
    )

    registry = OperasRegistry()
    loaded = load_user_ops(str(op_file), registry, namespace="ops")

    assert loaded == ["ops.scale_three"]
    declaration = registry.get("ops.scale_three")
    assert isinstance(declaration, OperaFunction)
    assert declaration.namespace == "ops"
    assert declaration.name == "scale_three"
    assert registry.call("ops.scale_three", 4) == 12
    info = registry.info("ops.scale_three")
    assert info["metadata"]["kind"] == "decl"
    assert info["metadata"]["source"] == "__JARVIS_OPERAS__"




from __future__ import annotations

import os

_DEV_WRITE_ENV = "JARVIS_OPERAS_DEV_WRITE"
_SUPPORTED_USER_WRITE_COMMANDS = frozenset({"load"})


def dev_write_enabled() -> bool:
    raw = os.getenv(_DEV_WRITE_ENV, "")
    normalized = raw.strip().lower()
    return normalized in {"1", "true", "yes", "on"}


def ensure_cli_write_allowed(command: str) -> None:
    normalized = str(command).strip().lower()
    if normalized in _SUPPORTED_USER_WRITE_COMMANDS:
        return
    if dev_write_enabled():
        return
    raise ValueError(
        f"CLI write command '{command}' is disabled by JO policy. "
        "Supported user write path is 'jopera load'. "
        f"For developer maintenance, set {_DEV_WRITE_ENV}=1."
    )




from __future__ import annotations

from concurrent.futures import Executor
from typing import Any

from .registry import OperasRegistry


class Operas:
    """Public function interface with sync + asyncio-compatible execution."""

    def __init__(
        self,
        registry: OperasRegistry | None = None,
        *,
        numpy_concurrency: int | None = None,
        executor: Executor | None = None,
    ) -> None:
        if numpy_concurrency is not None and numpy_concurrency < 1:
            raise ValueError("numpy_concurrency must be >= 1")
        if registry is None:
            if numpy_concurrency is None:
                self.registry = OperasRegistry(executor=executor)
            else:
                self.registry = OperasRegistry(
                    executor=executor,
                    numpy_concurrency=numpy_concurrency,
                )
            return

        self.registry = registry
        if numpy_concurrency is not None:
            self.registry.set_numpy_concurrency(numpy_concurrency)
        if executor is not None:
            self.registry.set_executor(executor)

    def call(
        self,
        name: str,
        *args: Any,
        backend: str = "numpy",
        **kwargs: Any,
    ) -> Any:
        return self.registry.call(name, *args, backend=backend, **kwargs)

    async def acall(
        self,
        name: str,
        *args: Any,
        backend: str = "numpy",
        **kwargs: Any,
    ) -> Any:
        return await self.registry.acall(name, *args, backend=backend, **kwargs)


def get_global_operas_registry() -> OperasRegistry:
    # Single global state is managed in jarvis_operas.api.
    from ..api import get_global_operas_registry as _get_global_operas_registry

    return _get_global_operas_registry()


def get_global_operas(*, numpy_concurrency: int | None = None) -> Operas:
    # Single global state is managed in jarvis_operas.api.
    from ..api import get_global_operas as _get_global_operas

    return _get_global_operas(numpy_concurrency=numpy_concurrency)



from __future__ import annotations

import asyncio
import time

import numpy as np
import pytest

from jarvis_operas import OperaFunction, Operas, OperasRegistry
from jarvis_operas.errors import OperatorCallError


def test_operas_numpy_backend_dispatch() -> None:
    registry = OperasRegistry()
    registry.register(
        OperaFunction(
            namespace="math",
            name="add_vec",
            arity=2,
            return_dtype=None,
            numpy_impl=lambda a, b: a + b,
        )
    )
    operas = Operas(registry, numpy_concurrency=2)

    left = np.array([1.0, 2.0, 3.0])
    right = np.array([4.0, 5.0, 6.0])
    value = operas.call("math.add_vec", left, right, backend="numpy")
    assert np.allclose(value, np.array([5.0, 7.0, 9.0]))


def test_operas_acall_numpy_backend_is_non_blocking() -> None:
    registry = OperasRegistry()

    def slow_add(a, b):
        time.sleep(0.2)
        return a + b

    registry.register(
        OperaFunction(
            namespace="math",
            name="slow_add",
            arity=2,
            return_dtype=None,
            numpy_impl=slow_add,
        )
    )
    operas = Operas(registry, numpy_concurrency=1)

    async def _runner() -> tuple[float, np.ndarray]:
        start = time.perf_counter()
        tick_task = asyncio.create_task(asyncio.sleep(0.02, result=True))
        call_task = asyncio.create_task(
            operas.acall(
                "math.slow_add",
                np.array([1.0, 2.0]),
                np.array([3.0, 4.0]),
                backend="numpy",
            )
        )
        await tick_task
        elapsed = time.perf_counter() - start
        value = await call_task
        return elapsed, value

    elapsed, value = asyncio.run(_runner())
    assert elapsed < 0.15
    assert np.allclose(value, np.array([4.0, 6.0]))


def test_operas_polars_native_expr_dispatch() -> None:
    pl = pytest.importorskip("polars")
    registry = OperasRegistry()

    def _should_not_run(*_args, **_kwargs):
        raise RuntimeError("numpy_impl should not be used for native polars dispatch")

    registry.register(
        OperaFunction(
            namespace="math",
            name="sum_native",
            arity=2,
            return_dtype=pl.Float64,
            numpy_impl=_should_not_run,
            polars_expr_impl=lambda x, y: x + y,
        )
    )

    operas = Operas(registry)
    expr = operas.call("math.sum_native", pl.col("a"), pl.col("b"), backend="polars")
    frame = pl.DataFrame({"a": [1.0, 2.0], "b": [3.0, 4.0]})
    values = frame.select(expr.alias("z")).get_column("z").to_list()
    assert values == [4.0, 6.0]


def test_operas_polars_fallback_map_batches_dispatch() -> None:
    pl = pytest.importorskip("polars")
    registry = OperasRegistry()

    registry.register(
        OperaFunction(
            namespace="math",
            name="sum_fallback",
            arity=2,
            return_dtype=pl.Float64,
            numpy_impl=lambda x, y: x + y,
            polars_expr_impl=None,
        )
    )

    operas = Operas(registry)
    expr = operas.call("math.sum_fallback", pl.col("a"), pl.col("b"), backend="polars")
    frame = pl.DataFrame({"a": [1.0, 2.0], "b": [3.0, 4.0]})
    values = frame.select(expr.alias("z")).get_column("z").to_list()
    assert values == [4.0, 6.0]


def test_operas_polars_fallback_requires_return_dtype() -> None:
    pl = pytest.importorskip("polars")
    registry = OperasRegistry()

    registry.register(
        OperaFunction(
            namespace="math",
            name="missing_dtype",
            arity=1,
            return_dtype=None,
            numpy_impl=lambda x: x,
            polars_expr_impl=None,
        )
    )

    operas = Operas(registry)
    with pytest.raises(OperatorCallError):
        operas.call("math.missing_dtype", pl.col("a"), backend="polars")



from __future__ import annotations

from typing import Any, Mapping

from .bootstrap_state import is_bootstrapping_global_registry
from .core.spec import OperaFunction
from .curve_cache import load_hot_curve_function_table
from .curve_manifest import (
    _build_curve_metadata,
    _normalize_selected_namespaces,
    _resolve_curve_bounds,
    _resolve_curve_namespace,
    _resolve_curve_return_dtype,
)
from .curve_storage import (
    _read_json_object,
    get_curve_index_path,
)
from .errors import OperatorConflict, OperatorNotFound
from .logging import get_logger
from .name_utils import try_split_full_name


def _build_existing_interpolation_name_index(registry: Any) -> dict[str, list[str]]:
    """Build short-name index for already-registered interpolation declarations."""

    indexed: dict[str, list[str]] = {}
    for existing_name in registry.list():
        split = try_split_full_name(existing_name)
        if split is None:
            continue
        _, short_name = split
        try:
            declaration = registry.get(existing_name)
        except OperatorNotFound:
            continue

        metadata = declaration.metadata
        if not isinstance(metadata, Mapping):
            continue
        if metadata.get("category") != "interpolation":
            continue
        indexed.setdefault(short_name, []).append(existing_name)
    return indexed


def register_hot_curves_in_registry(
    registry: Any,
    *,
    namespace: str = "interp",
    cache_root: str | None = None,
    index_path: str | None = None,
    namespaces: list[str] | tuple[str, ...] | None = None,
    include_cold: bool = False,
    overwrite: bool = True,
    logger: Any | None = None,
) -> list[str]:
    """Register hot curves into a function namespace."""

    local_logger = get_logger(logger, action="register_hot_curves_in_registry")
    resolved_index = get_curve_index_path(cache_root=cache_root, index_path=index_path)
    if not resolved_index.exists():
        return []

    raw_index = _read_json_object(resolved_index)
    raw_curves = raw_index.get("curves", {})
    if not isinstance(raw_curves, Mapping):
        raise ValueError(f"invalid curve index format: {resolved_index}")

    selected_namespaces = _normalize_selected_namespaces(namespaces)
    loaded = load_hot_curve_function_table(
        cache_root=cache_root,
        index_path=index_path,
        namespaces=namespaces,
        include_cold=include_cold,
        logger=logger,
    )
    existing_interpolation_names = (
        _build_existing_interpolation_name_index(registry) if overwrite else {}
    )
    registered: list[str] = []
    for curve_id, curve_fn in loaded.items():
        curve_entry = raw_curves.get(curve_id, {})
        target_namespace = _resolve_curve_namespace(
            curve_entry=curve_entry if isinstance(curve_entry, Mapping) else {},
            default_namespace=namespace,
        )
        resolved_x_min, resolved_x_max = _resolve_curve_bounds(
            curve_entry=curve_entry if isinstance(curve_entry, Mapping) else {},
            curve_fn=curve_fn,
        )
        metadata_override = (
            curve_entry.get("metadata")
            if isinstance(curve_entry, Mapping) and isinstance(curve_entry.get("metadata"), Mapping)
            else None
        )
        metadata = _build_curve_metadata(
            curve_id=curve_id,
            target_namespace=target_namespace,
            source=(
                str(curve_entry.get("source"))
                if isinstance(curve_entry, Mapping) and curve_entry.get("source") is not None
                else None
            ),
            kind=(
                str(curve_entry.get("kind"))
                if isinstance(curve_entry, Mapping) and curve_entry.get("kind") is not None
                else None
            ),
            log_x=bool(curve_entry.get("logX", False)) if isinstance(curve_entry, Mapping) else False,
            log_y=bool(curve_entry.get("logY", False)) if isinstance(curve_entry, Mapping) else False,
            hot=bool(curve_entry.get("hot", True)) if isinstance(curve_entry, Mapping) else True,
            points=curve_entry.get("points") if isinstance(curve_entry, Mapping) else None,
            x_min=resolved_x_min,
            x_max=resolved_x_max,
            metadata_override=metadata_override,
        )

        if selected_namespaces is not None and target_namespace not in selected_namespaces:
            continue

        full_name = f"{target_namespace}.{curve_id}"
        if overwrite:
            for existing_name in existing_interpolation_names.get(curve_id, []):
                try:
                    registry.delete(existing_name)
                except OperatorNotFound:
                    continue
            existing_interpolation_names[curve_id] = []

        declaration = OperaFunction(
            namespace=target_namespace,
            name=curve_id,
            arity=1,
            return_dtype=_resolve_curve_return_dtype(curve_entry),
            numpy_impl=curve_fn,
            metadata=metadata,
        )
        try:
            registry.register(declaration)
            registered.append(full_name)
            if overwrite:
                existing_interpolation_names[curve_id] = [full_name]
        except OperatorConflict:
            if overwrite:
                raise
            continue

    local_logger.info(
        "registered {} interpolation operators",
        len(registered),
    )
    if not is_bootstrapping_global_registry():
        from .integration import refresh_sympy_dicts_if_global_registry

        refresh_sympy_dicts_if_global_registry(registry)
    return sorted(registered)


def register_hot_curve_namespace_in_registry(
    registry: Any,
    namespace: str,
    *,
    cache_root: str | None = None,
    index_path: str | None = None,
    include_cold: bool = False,
    overwrite: bool = True,
    logger: Any | None = None,
) -> list[str]:
    """Register hot interpolation functions for one namespace."""

    return register_hot_curves_in_registry(
        registry,
        namespace=namespace,
        cache_root=cache_root,
        index_path=index_path,
        namespaces=[namespace],
        include_cold=include_cold,
        overwrite=overwrite,
        logger=logger,
    )



from __future__ import annotations

import json
import pickle
from concurrent.futures import ThreadPoolExecutor
from importlib import resources
from pathlib import Path
from typing import Any

import numpy as np
import pytest

from jarvis_operas import OperaFunction, OperasRegistry, func_locals, numeric_funcs
from jarvis_operas.api import get_global_operas_registry
from jarvis_operas.curves import (
    add_interpolation_namespace,
    Curve1DInterpolator,
    get_curve_index_path,
    init_curve_cache,
    interpolation_manifest_resource,
    list_interpolation_namespace_entries,
    list_interpolation_namespaces,
    load_hot_curve_function_table,
    load_hot_curve_namespace_function_table,
    load_interpolation_manifest_library,
    remove_interpolation_namespace,
    register_hot_curves,
    register_hot_curve_namespace_in_registry,
    register_hot_curves_in_registry,
    validate_interpolation_namespace_index,
)
from jarvis_operas.integration import refresh_sympy_dicts_if_global_registry


def _write_curve(path, x_values, y_values) -> None:
    path.write_text(
        json.dumps({"x": list(x_values), "y": list(y_values)}),
        encoding="utf-8",
    )


def test_curve_interpolator_scalar_and_vector_calls_are_consistent() -> None:
    curve = Curve1DInterpolator(
        curve_id="flat_linear",
        x_values=np.array([1.0, 2.0, 3.0]),
        y_values=np.array([10.0, 20.0, 30.0]),
        kind="linear",
        log_x=False,
        log_y=False,
    )

    scalar = curve(2.5)
    vector = curve(np.array([2.5, 3.0]))

    assert isinstance(scalar, float)
    assert np.isclose(scalar, 25.0)
    assert isinstance(vector, np.ndarray)
    assert np.allclose(vector, np.array([25.0, 30.0]))


def test_curve_interpolator_clip_mode_applies_before_logx() -> None:
    curve = Curve1DInterpolator(
        curve_id="clip_logx",
        x_values=np.array([1.0, 2.0, 4.0, 8.0]),
        y_values=np.array([1.0, 4.0, 16.0, 64.0]),
        kind="linear",
        log_x=True,
        log_y=False,
        extrapolation="clip",
    )

    baseline = curve(1.0)
    assert np.isclose(curve(-10.0), baseline)
    assert np.isclose(curve(np.array([-10.0]))[0], baseline)


def test_curve_interpolator_pickle_roundtrip_keeps_call_behavior() -> None:
    curve = Curve1DInterpolator(
        curve_id="pickle_roundtrip",
        x_values=np.array([1.0, 2.0, 4.0, 8.0]),
        y_values=np.array([1.0, 2.0, 4.0, 8.0]),
        kind="linear",
        log_x=True,
        log_y=True,
        extrapolation="extrapolate",
    )
    restored = pickle.loads(pickle.dumps(curve, protocol=5))

    assert np.isclose(restored(3.0), curve(3.0))
    assert np.allclose(restored(np.array([2.0, 3.0, 5.0])), curve(np.array([2.0, 3.0, 5.0])))


def test_curve_interpolator_nan_mode_returns_nan_for_nan_and_out_of_bounds() -> None:
    curve = Curve1DInterpolator(
        curve_id="nan_mode",
        x_values=np.array([1.0, 2.0, 4.0, 8.0]),
        y_values=np.array([1.0, 2.0, 4.0, 8.0]),
        kind="linear",
        log_x=True,
        log_y=True,
        extrapolation="nan",
    )

    assert np.isnan(curve(np.nan))
    assert np.isnan(curve(0.5))
    values = curve(np.array([np.nan, 0.5, 2.0]))
    assert values.shape == (3,)
    assert np.isnan(values[0])
    assert np.isnan(values[1])
    assert np.isfinite(values[2])


def test_init_curve_cache_and_load_hot_functions(tmp_path) -> None:
    source_dir = tmp_path / "source"
    source_dir.mkdir()
    _write_curve(source_dir / "hot.json", [0.0, 1.0, 2.0], [0.0, 1.0, 2.0])
    _write_curve(source_dir / "cold.json", [0.0, 1.0, 2.0], [0.0, 2.0, 4.0])

    manifest_path = tmp_path / "manifest.json"
    manifest_path.write_text(
        json.dumps(
            {
                "curves": [
                    {
                        "curve_id": "line_hot",
                        "source": "source/hot.json",
                        "kind": "linear",
                        "hot": True,
                    },
                    {
                        "curve_id": "line_cold",
                        "source": "source/cold.json",
                        "kind": "linear",
                        "hot": False,
                    },
                ]
            }
        ),
        encoding="utf-8",
    )

    cache_root = tmp_path / "cache"
    summary = init_curve_cache(str(manifest_path), cache_root=str(cache_root))
    assert summary["total"] == 2
    assert sorted(summary["compiled"]) == ["line_cold", "line_hot"]
    assert get_curve_index_path(cache_root=str(cache_root)).exists()

    hot_table = load_hot_curve_function_table(cache_root=str(cache_root))
    assert set(hot_table.keys()) == {"line_hot"}
    assert np.isclose(hot_table["line_hot"](1.5), 1.5)

    full_table = load_hot_curve_function_table(cache_root=str(cache_root), include_cold=True)
    assert set(full_table.keys()) == {"line_hot", "line_cold"}
    assert np.isclose(full_table["line_cold"](1.5), 3.0)


def test_init_curve_cache_reuses_existing_pickle(tmp_path) -> None:
    source_dir = tmp_path / "source"
    source_dir.mkdir()
    _write_curve(source_dir / "curve.json", [0.0, 1.0, 2.0], [0.0, 1.0, 4.0])

    manifest_path = tmp_path / "manifest.json"
    manifest_path.write_text(
        json.dumps(
            {
                "curves": [
                    {
                        "curve_id": "quadratic",
                        "source": "source/curve.json",
                        "kind": "linear",
                    }
                ]
            }
        ),
        encoding="utf-8",
    )

    cache_root = tmp_path / "cache"
    first = init_curve_cache(str(manifest_path), cache_root=str(cache_root))
    second = init_curve_cache(str(manifest_path), cache_root=str(cache_root))

    assert first["compiled"] == ["quadratic"]
    assert second["compiled"] == []
    assert second["cached"] == ["quadratic"]


def test_register_hot_curves_updates_function_table(tmp_path) -> None:
    source_dir = tmp_path / "source"
    source_dir.mkdir()
    _write_curve(source_dir / "curve.json", [0.0, 1.0, 2.0], [1.0, 2.0, 3.0])

    manifest_path = tmp_path / "manifest.json"
    manifest_path.write_text(
        json.dumps(
            {
                "curves": [
                    {
                        "curve_id": "offset_line",
                        "source": "source/curve.json",
                        "kind": "linear",
                        "hot": True,
                    }
                ]
            }
        ),
        encoding="utf-8",
    )
    cache_root = tmp_path / "cache"
    init_curve_cache(str(manifest_path), cache_root=str(cache_root))

    function_table: dict[str, object] = {}
    updated = register_hot_curves(function_table, cache_root=str(cache_root))

    assert updated == ["offset_line"]
    assert "offset_line" in function_table
    assert callable(function_table["offset_line"])


def test_packaged_interpolation_manifest_library_is_available() -> None:
    payload = load_interpolation_manifest_library()
    assert payload["kind"] == "jarvis_operas_interpolation_namespace_index"
    assert isinstance(payload["namespaces"], list)
    assert isinstance(payload["curves"], list)
    assert interpolation_manifest_resource().endswith("manifests/interpolations.manifest.json")


def test_packaged_interpolation_manifest_sources_exist() -> None:
    payload = load_interpolation_manifest_library()
    base = resources.files("jarvis_operas").joinpath("manifests")
    for item in payload["curves"]:
        source = item.get("source")
        assert isinstance(source, str) and source.strip()
        path = base.joinpath(Path(source))
        assert path.is_file()


def test_list_interpolation_namespaces_works() -> None:
    namespaces = list_interpolation_namespaces()
    assert "interp1" in namespaces
    assert "dmdd" in namespaces


def test_init_curve_cache_can_filter_namespaces(tmp_path) -> None:
    base = resources.files("jarvis_operas").joinpath("manifests")
    manifest_path = base.joinpath("interpolations.manifest.json")

    index_path = tmp_path / "cache" / "index.json"

    summary = init_curve_cache(
        str(manifest_path),
        index_path=str(index_path),
        namespaces=["dmdd"],
        force=True,
    )
    assert summary["total"] == 1
    table = load_hot_curve_function_table(index_path=str(index_path))
    assert set(table.keys()) == {"LZSI2024"}


def test_register_hot_curves_in_registry_uses_group_namespace_and_updates(tmp_path) -> None:
    source_dir = tmp_path / "source"
    source_dir.mkdir()
    _write_curve(source_dir / "curve.json", [1.0, 2.0, 3.0, 4.0], [1.0, 2.0, 3.0, 4.0])

    manifest_path = tmp_path / "manifest.json"
    manifest_path.write_text(
        json.dumps(
            {
                "curves": [
                    {
                        "curve_id": "LZSI2024",
                        "source": "source/curve.json",
                        "kind": "cubic",
                        "logX": True,
                        "logY": True,
                        "hot": True,
                        "metadata": {"group": "interp"},
                    }
                ]
            }
        ),
        encoding="utf-8",
    )
    cache_root = tmp_path / "cache"
    init_curve_cache(str(manifest_path), cache_root=str(cache_root))

    registry = OperasRegistry()
    first = register_hot_curves_in_registry(registry, cache_root=str(cache_root))
    assert first == ["interp.LZSI2024"]
    assert "interp.LZSI2024" in registry.list(namespace="interp")

    # Move group to dmdd and re-init; registry update should replace old namespace item.
    manifest_path.write_text(
        json.dumps(
            {
                "curves": [
                    {
                        "curve_id": "LZSI2024",
                        "source": "source/curve.json",
                        "kind": "cubic",
                        "logX": True,
                        "logY": True,
                        "hot": True,
                        "metadata": {"group": "dmdd"},
                    }
                ]
            }
        ),
        encoding="utf-8",
    )
    init_curve_cache(str(manifest_path), cache_root=str(cache_root))
    second = register_hot_curves_in_registry(registry, cache_root=str(cache_root))
    assert second == ["dmdd.LZSI2024"]
    assert "dmdd.LZSI2024" in registry.list(namespace="dmdd")
    assert registry.list(namespace="interp") == []


def test_register_hot_curves_in_registry_overwrite_keeps_non_interpolation_same_short_name(
    tmp_path,
) -> None:
    source_dir = tmp_path / "source"
    source_dir.mkdir()
    _write_curve(source_dir / "curve.json", [1.0, 2.0, 3.0], [1.0, 2.0, 3.0])

    manifest_path = tmp_path / "manifest.json"
    manifest_path.write_text(
        json.dumps(
            {
                "curves": [
                    {
                        "curve_id": "LZSI2024",
                        "source": "source/curve.json",
                        "kind": "linear",
                        "hot": True,
                        "metadata": {"group": "dmdd"},
                    }
                ]
            }
        ),
        encoding="utf-8",
    )
    cache_root = tmp_path / "cache"
    init_curve_cache(str(manifest_path), cache_root=str(cache_root))

    registry = OperasRegistry()
    registry.register(
        OperaFunction(
            namespace="user_ops",
            name="LZSI2024",
            arity=1,
            return_dtype=None,
            numpy_impl=lambda x: x,
            metadata={"category": "user_defined"},
        )
    )

    loaded = register_hot_curves_in_registry(
        registry,
        cache_root=str(cache_root),
        overwrite=True,
    )

    assert loaded == ["dmdd.LZSI2024"]
    assert "dmdd.LZSI2024" in registry.list(namespace="dmdd")
    assert "user_ops.LZSI2024" in registry.list(namespace="user_ops")


def test_register_hot_curves_in_global_registry_refreshes_sympy_dicts(tmp_path) -> None:
    source_dir = tmp_path / "source"
    source_dir.mkdir()
    _write_curve(source_dir / "curve.json", [1.0, 2.0, 3.0, 4.0], [2.0, 3.0, 4.0, 5.0])

    manifest_path = tmp_path / "manifest.json"
    manifest_path.write_text(
        json.dumps(
            {
                "curves": [
                    {
                        "curve_id": "AutoRefreshCurve",
                        "source": "source/curve.json",
                        "kind": "cubic",
                        "hot": True,
                        "metadata": {"group": "dmdd_refresh"},
                    }
                ]
            }
        ),
        encoding="utf-8",
    )

    index_path = tmp_path / "cache" / "index.json"
    init_curve_cache(str(manifest_path), index_path=str(index_path))

    registry = get_global_operas_registry()
    registry.delete_namespace("dmdd_refresh")
    refresh_sympy_dicts_if_global_registry(registry)
    full_name = "dmdd_refresh.AutoRefreshCurve"
    assert "dmdd_refresh" not in func_locals

    try:
        loaded = register_hot_curves_in_registry(registry, index_path=str(index_path))
        assert full_name in loaded
        assert "dmdd_refresh" in func_locals
        symbol_name = str(getattr(func_locals["dmdd_refresh"], "AutoRefreshCurve"))
        assert symbol_name in numeric_funcs
    finally:
        registry.delete_namespace("dmdd_refresh")
        refresh_sympy_dicts_if_global_registry(registry)


def test_register_hot_curves_in_registry_can_filter_namespaces(tmp_path) -> None:
    source_dir = tmp_path / "source"
    source_dir.mkdir()
    _write_curve(source_dir / "curve_a.json", [1.0, 2.0, 3.0], [1.0, 2.0, 3.0])
    _write_curve(source_dir / "curve_b.json", [1.0, 2.0, 3.0], [2.0, 3.0, 4.0])

    manifest_path = tmp_path / "manifest.json"
    manifest_path.write_text(
        json.dumps(
            {
                "version": 2,
                "kind": "jarvis_operas_interpolation_namespace_index",
                "namespaces": [
                    {
                        "namespace": "ns1",
                        "manifest": "ns1.json",
                    },
                    {
                        "namespace": "ns2",
                        "manifest": "ns2.json",
                    },
                ],
            }
        ),
        encoding="utf-8",
    )
    (tmp_path / "ns1.json").write_text(
        json.dumps(
            {
                "namespace": "ns1",
                "functions": [
                    {
                        "name": "f1",
                        "source": "source/curve_a.json",
                        "kind": "linear",
                        "hot": True,
                    }
                ],
            }
        ),
        encoding="utf-8",
    )
    (tmp_path / "ns2.json").write_text(
        json.dumps(
            {
                "namespace": "ns2",
                "functions": [
                    {
                        "name": "f2",
                        "source": "source/curve_b.json",
                        "kind": "linear",
                        "hot": True,
                    }
                ],
            }
        ),
        encoding="utf-8",
    )
    index_path = tmp_path / "cache" / "index.json"
    init_curve_cache(str(manifest_path), index_path=str(index_path))

    registry = OperasRegistry()
    loaded = register_hot_curves_in_registry(
        registry,
        index_path=str(index_path),
        namespaces=["ns2"],
    )
    assert loaded == ["ns2.f2"]
    assert registry.list(namespace="ns1") == []
    assert registry.list(namespace="ns2") == ["ns2.f2"]


def test_namespace_specific_curve_helpers(tmp_path) -> None:
    source_dir = tmp_path / "source"
    source_dir.mkdir()
    _write_curve(source_dir / "curve_a.json", [0.0, 1.0], [0.0, 1.0])
    _write_curve(source_dir / "curve_b.json", [0.0, 1.0], [1.0, 2.0])

    manifest_path = tmp_path / "manifest.json"
    manifest_path.write_text(
        json.dumps(
            {
                "version": 2,
                "kind": "jarvis_operas_interpolation_namespace_index",
                "namespaces": [
                    {"namespace": "alpha", "manifest": "alpha.json"},
                    {"namespace": "beta", "manifest": "beta.json"},
                ],
            }
        ),
        encoding="utf-8",
    )
    (tmp_path / "alpha.json").write_text(
        json.dumps(
            {
                "namespace": "alpha",
                "functions": [
                    {"name": "f1", "source": "source/curve_a.json", "kind": "linear", "hot": True}
                ],
            }
        ),
        encoding="utf-8",
    )
    (tmp_path / "beta.json").write_text(
        json.dumps(
            {
                "namespace": "beta",
                "functions": [
                    {"name": "f2", "source": "source/curve_b.json", "kind": "linear", "hot": True}
                ],
            }
        ),
        encoding="utf-8",
    )
    index_path = tmp_path / "cache" / "index.json"
    init_curve_cache(str(manifest_path), index_path=str(index_path))

    alpha_table = load_hot_curve_namespace_function_table("alpha", index_path=str(index_path))
    beta_table = load_hot_curve_namespace_function_table("beta", index_path=str(index_path))
    assert set(alpha_table.keys()) == {"f1"}
    assert set(beta_table.keys()) == {"f2"}

    registry = OperasRegistry()
    loaded = register_hot_curve_namespace_in_registry(
        registry,
        "beta",
        index_path=str(index_path),
    )
    assert loaded == ["beta.f2"]
    assert registry.list(namespace="alpha") == []
    assert registry.list(namespace="beta") == ["beta.f2"]


def test_interpolation_namespace_entry_management_and_validation(tmp_path) -> None:
    source_dir = tmp_path / "source"
    source_dir.mkdir()
    _write_curve(source_dir / "a.json", [0.0, 1.0], [0.0, 1.0])
    _write_curve(source_dir / "b.json", [0.0, 1.0], [1.0, 2.0])
    _write_curve(source_dir / "c.json", [0.0, 1.0], [2.0, 3.0])

    root_manifest = tmp_path / "interpolations.manifest.json"
    root_manifest.write_text(
        json.dumps(
            {
                "version": 2,
                "kind": "jarvis_operas_interpolation_namespace_index",
                "namespaces": [
                    {"namespace": "alpha", "manifest": "alpha.manifest.json"}
                ],
            }
        ),
        encoding="utf-8",
    )
    (tmp_path / "alpha.manifest.json").write_text(
        json.dumps(
            {
                "namespace": "alpha",
                "functions": [
                    {"name": "a1", "source": "source/a.json", "kind": "linear", "hot": True}
                ],
            }
        ),
        encoding="utf-8",
    )
    (tmp_path / "beta.manifest.json").write_text(
        json.dumps(
            {
                "namespace": "beta",
                "functions": [
                    {"name": "b1", "source": "source/b.json", "kind": "linear", "hot": True}
                ],
            }
        ),
        encoding="utf-8",
    )
    (tmp_path / "beta2.manifest.json").write_text(
        json.dumps(
            {
                "namespace": "beta",
                "functions": [
                    {"name": "b2", "source": "source/c.json", "kind": "linear", "hot": True}
                ],
            }
        ),
        encoding="utf-8",
    )

    add_result = add_interpolation_namespace(
        manifest_path=str(root_manifest),
        namespace="beta",
        namespace_manifest="beta.manifest.json",
        description="beta namespace",
    )
    assert add_result["added"] is True
    assert add_result["updated"] is False
    assert add_result["namespace"] == "beta"

    unchanged = add_interpolation_namespace(
        manifest_path=str(root_manifest),
        namespace="beta",
        namespace_manifest="beta.manifest.json",
        description="beta namespace",
    )
    assert unchanged["unchanged"] is True

    with pytest.raises(ValueError):
        add_interpolation_namespace(
            manifest_path=str(root_manifest),
            namespace="beta",
            namespace_manifest="beta2.manifest.json",
        )

    updated = add_interpolation_namespace(
        manifest_path=str(root_manifest),
        namespace="beta",
        namespace_manifest="beta2.manifest.json",
        overwrite=True,
    )
    assert updated["updated"] is True

    entries = list_interpolation_namespace_entries(manifest_path=str(root_manifest))
    assert [item["namespace"] for item in entries] == ["alpha", "beta"]
    assert entries[1]["manifest"] == "beta2.manifest.json"

    report = validate_interpolation_namespace_index(manifest_path=str(root_manifest))
    assert report["ok"] is True
    assert report["namespace_count"] == 2
    assert report["function_count"] == 2

    remove_result = remove_interpolation_namespace(
        manifest_path=str(root_manifest),
        namespace="beta",
    )
    assert remove_result["removed"] is True
    assert remove_result["remaining"] == 1

    missing_remove = remove_interpolation_namespace(
        manifest_path=str(root_manifest),
        namespace="beta",
    )
    assert missing_remove["removed"] is False


def test_interpolation_function_supports_async_and_polars_backend(tmp_path) -> None:
    pl = pytest.importorskip("polars")
    source_dir = tmp_path / "source"
    source_dir.mkdir()
    _write_curve(source_dir / "curve.json", [0.0, 1.0, 2.0], [0.0, 1.0, 2.0])

    manifest_path = tmp_path / "manifest.json"
    manifest_path.write_text(
        json.dumps(
            {
                "curves": [
                    {
                        "curve_id": "line",
                        "source": "source/curve.json",
                        "kind": "linear",
                        "hot": True,
                        "metadata": {"group": "interp"},
                    }
                ]
            }
        ),
        encoding="utf-8",
    )
    index_path = tmp_path / "cache" / "index.json"
    init_curve_cache(str(manifest_path), index_path=str(index_path))

    registry = OperasRegistry()
    register_hot_curves_in_registry(registry, index_path=str(index_path))

    sync_value = registry.call("interp.line", x=np.array([0.5, 1.5]))
    assert np.allclose(sync_value, np.array([0.5, 1.5]))

    import asyncio

    async_value = asyncio.run(registry.acall("interp.line", x=np.array([0.25, 1.25])))
    assert np.allclose(async_value, np.array([0.25, 1.25]))

    expr = registry.call("interp.line", pl.col("x"), backend="polars")
    frame = pl.DataFrame({"x": [0.0, 1.0, 2.0]})
    values = frame.select(expr.alias("y")).get_column("y").to_list()
    assert values == [0.0, 1.0, 2.0]


def test_interpolation_info_note_contains_input_range(tmp_path) -> None:
    source_dir = tmp_path / "source"
    source_dir.mkdir()
    _write_curve(source_dir / "curve.json", [0.0, 1.0, 2.0], [0.0, 1.0, 2.0])

    manifest_path = tmp_path / "manifest.json"
    manifest_path.write_text(
        json.dumps(
            {
                "curves": [
                    {
                        "curve_id": "line_note",
                        "source": "source/curve.json",
                        "kind": "linear",
                        "hot": True,
                        "metadata": {
                            "group": "interp",
                            "note": "Custom interpolation note.",
                        },
                    }
                ]
            }
        ),
        encoding="utf-8",
    )
    index_path = tmp_path / "cache" / "index.json"
    init_curve_cache(str(manifest_path), index_path=str(index_path))

    registry = OperasRegistry()
    register_hot_curves_in_registry(registry, index_path=str(index_path))
    info = registry.info("interp.line_note")
    note = info["metadata"].get("note")
    assert isinstance(note, str)
    assert note == "Custom interpolation note.\n\tInput range: x in [0, 2]."


def test_init_curve_cache_concurrent_writes_are_atomic(tmp_path) -> None:
    source_dir = tmp_path / "source"
    source_dir.mkdir()
    _write_curve(source_dir / "curve.json", [0.0, 1.0, 2.0], [0.0, 1.0, 2.0])

    manifest_path = tmp_path / "manifest.json"
    manifest_path.write_text(
        json.dumps(
            {
                "curves": [
                    {
                        "curve_id": "line",
                        "source": "source/curve.json",
                        "kind": "linear",
                        "hot": True,
                        "metadata": {"group": "interp"},
                    }
                ]
            }
        ),
        encoding="utf-8",
    )
    index_path = tmp_path / "cache" / "index.json"

    def _run_once(_: int) -> dict[str, Any]:
        return init_curve_cache(
            str(manifest_path),
            index_path=str(index_path),
            force=True,
        )

    with ThreadPoolExecutor(max_workers=8) as pool:
        results = list(pool.map(_run_once, range(16)))

    assert results
    assert index_path.exists()
    table = load_hot_curve_function_table(index_path=str(index_path))
    assert "line" in table
    assert np.isclose(float(table["line"](1.5)), 1.5)

    temp_files = list(index_path.parent.glob("*.tmp"))
    assert temp_files == []


def test_registered_curve_reuses_prepared_interpolator(tmp_path) -> None:
    source_dir = tmp_path / "source"
    source_dir.mkdir()
    _write_curve(source_dir / "curve.json", [0.0, 1.0, 2.0], [0.0, 1.0, 2.0])

    manifest_path = tmp_path / "manifest.json"
    manifest_path.write_text(
        json.dumps(
            {
                "curves": [
                    {
                        "curve_id": "line",
                        "source": "source/curve.json",
                        "kind": "linear",
                        "hot": True,
                        "metadata": {"group": "interp"},
                    }
                ]
            }
        ),
        encoding="utf-8",
    )
    index_path = tmp_path / "cache" / "index.json"
    init_curve_cache(str(manifest_path), index_path=str(index_path), force=True)

    registry = OperasRegistry()
    register_hot_curves_in_registry(registry, index_path=str(index_path))
    declaration = registry.get("interp.line")
    curve_fn = declaration.numpy_impl
    assert isinstance(curve_fn, Curve1DInterpolator)
    interp_id = id(curve_fn._interp)

    curve_fn._build_interp = lambda: (_ for _ in ()).throw(AssertionError("should not rebuild"))  # type: ignore[method-assign]

    first = registry.call("interp.line", x=0.5)
    second = registry.call("interp.line", x=np.array([0.5, 1.5]))
    assert np.isclose(float(first), 0.5)
    assert np.allclose(second, np.array([0.5, 1.5]))
    assert id(curve_fn._interp) == interp_id
    assert registry.get("interp.line").numpy_impl is curve_fn


def test_dmdd_curve_returns_nan_on_out_of_bounds_and_nan_input() -> None:
    base = resources.files("jarvis_operas").joinpath("manifests")
    manifest_path = base.joinpath("interpolations.manifest.json")

    # Compile against current packaged manifests to avoid stale local cache behavior.
    temp_index = Path.cwd() / ".pytest-dmdd-index.json"
    try:
        init_curve_cache(
            str(manifest_path),
            index_path=str(temp_index),
            namespaces=["dmdd"],
            force=True,
        )

        registry = OperasRegistry()
        loaded = register_hot_curves_in_registry(registry, index_path=str(temp_index))
        assert "dmdd.LZSI2024" in loaded

        scalar = registry.call("dmdd.LZSI2024", x=1.0)
        assert np.isnan(float(scalar))

        mixed = registry.call("dmdd.LZSI2024", x=np.array([1.0, np.nan, 1.0e3]))
        assert mixed.shape == (3,)
        assert np.isnan(mixed[0])
        assert np.isnan(mixed[1])
        assert np.isfinite(mixed[2])
    finally:
        if temp_index.exists():
            temp_index.unlink()
        temp_curve_dir = temp_index.parent / "curves"
        if temp_curve_dir.exists():
            for child in temp_curve_dir.glob("*"):
                child.unlink()
            temp_curve_dir.rmdir()


def test_global_registry_bootstrap_refreshes_builtin_curve_cache(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    cache_root = tmp_path / "curve-cache"
    monkeypatch.setenv("JARVIS_OPERAS_CURVE_CACHE_ROOT", str(cache_root))

    manifests_resource = resources.files("jarvis_operas").joinpath("manifests")
    with resources.as_file(manifests_resource) as manifests_dir:
        stale_namespace_manifest = tmp_path / "stale-dmdd.manifest.json"
        stale_namespace_manifest.write_text(
            json.dumps(
                {
                    "version": 1,
                    "kind": "jarvis_operas_interpolation_namespace_manifest",
                    "namespace": "dmdd",
                    "functions": [
                        {
                            "name": "LZSI2024",
                            "source": str((manifests_dir / "dmdd" / "LZSI_2024.json").resolve()),
                            "kind": "cubic",
                            "logX": True,
                            "logY": True,
                            "extrapolation": "extrapolate",
                            "hot": True,
                        }
                    ],
                }
            ),
            encoding="utf-8",
        )
        stale_root_manifest = tmp_path / "stale.manifest.json"
        stale_root_manifest.write_text(
            json.dumps(
                {
                    "version": 2,
                    "kind": "jarvis_operas_interpolation_namespace_index",
                    "namespaces": [
                        {
                            "namespace": "dmdd",
                            "manifest": str(stale_namespace_manifest.resolve()),
                        }
                    ],
                }
            ),
            encoding="utf-8",
        )
        init_curve_cache(str(stale_root_manifest), force=True)

    stale_table = load_hot_curve_function_table(cache_root=str(cache_root))
    assert "LZSI2024" in stale_table
    assert np.isinf(float(stale_table["LZSI2024"](1.0)))

    import jarvis_operas.api as api_mod

    monkeypatch.setattr(api_mod, "_global_operas_registry", None)
    monkeypatch.setattr(api_mod, "_global_operas", None)
    registry = api_mod.get_global_operas_registry()

    value = registry.call("dmdd.LZSI2024", x=1.0)
    assert np.isnan(float(value))



from __future__ import annotations

from .index import NAMESPACE_DECLARATIONS, NAMESPACE_LOADERS, get_catalog_declarations

__all__ = ["NAMESPACE_DECLARATIONS", "NAMESPACE_LOADERS", "get_catalog_declarations"]



from __future__ import annotations

from .operas import Operas, get_global_operas, get_global_operas_registry
from .registry import OperasRegistry
from .spec import OperaFunction

__all__ = [
    "OperaFunction",
    "OperasRegistry",
    "Operas",
    "get_global_operas_registry",
    "get_global_operas",
]



from __future__ import annotations

import inspect
from dataclasses import dataclass, field
from types import MappingProxyType
from typing import Any, Callable, Mapping

from ..errors import OperatorCallError
from ..name_utils import (
    normalize_namespace,
    normalize_short_name,
)


def _normalize_arity(arity: int | None) -> int | None:
    if arity is None:
        return None
    if not isinstance(arity, int):
        raise TypeError("OperaFunction.arity must be int or None")
    if arity < 0:
        raise ValueError("OperaFunction.arity must be >= 0")
    return arity


@dataclass(frozen=True)
class OperaFunction:
    """Unified declaration class for one operas function."""

    namespace: str
    name: str
    arity: int | None
    return_dtype: Any | None
    numpy_impl: Callable[..., Any] | None
    polars_expr_impl: Callable[..., Any] | None = None
    flags: frozenset[str] = field(default_factory=frozenset)
    metadata: Mapping[str, Any] = field(default_factory=dict)

    def __post_init__(self) -> None:
        object.__setattr__(self, "namespace", normalize_namespace(self.namespace))
        object.__setattr__(
            self,
            "name",
            normalize_short_name(self.name, field_label="OperaFunction.name"),
        )
        object.__setattr__(self, "arity", _normalize_arity(self.arity))

        if self.numpy_impl is not None and not callable(self.numpy_impl):
            raise TypeError("OperaFunction.numpy_impl must be callable or None")
        if self.polars_expr_impl is not None and not callable(self.polars_expr_impl):
            raise TypeError("OperaFunction.polars_expr_impl must be callable or None")
        if self.numpy_impl is None and self.polars_expr_impl is None:
            raise ValueError("OperaFunction requires at least one implementation")

        if isinstance(self.flags, frozenset):
            normalized_flags = self.flags
        else:
            normalized_flags = frozenset(str(flag) for flag in self.flags)
        object.__setattr__(self, "flags", normalized_flags)

        object.__setattr__(self, "metadata", MappingProxyType(dict(self.metadata or {})))

    @property
    def full_name(self) -> str:
        return f"{self.namespace}.{self.name}"

    @property
    def supports_numpy(self) -> bool:
        return self.numpy_impl is not None

    @property
    def supports_polars_native(self) -> bool:
        return self.polars_expr_impl is not None

    @property
    def supports_polars_fallback(self) -> bool:
        return self.numpy_impl is not None and self.return_dtype is not None

    def validate_arity(self, args: tuple[Any, ...]) -> None:
        if self.arity is None:
            return
        # Keep kwargs-friendly call style: strict arity check only when positional
        # arguments are used.
        if args and len(args) != self.arity:
            raise OperatorCallError(
                self.full_name,
                f"Operator '{self.full_name}' expects arity={self.arity}, got {len(args)} positional args.",
            )

    def dispatch(
        self,
        *args: Any,
        backend: str = "numpy",
        **kwargs: Any,
    ) -> Any:
        self.validate_arity(args)

        if backend == "numpy":
            if self.numpy_impl is None:
                raise OperatorCallError(
                    self.full_name,
                    f"Operator '{self.full_name}' does not provide numpy_impl.",
                )
            return self.numpy_impl(*args, **kwargs)

        if backend == "polars":
            from ..adapters.polars import dispatch_polars_expr

            return dispatch_polars_expr(
                operator_name=self.full_name,
                declaration=self,
                args=args,
                kwargs=kwargs,
            )

        raise OperatorCallError(
            self.full_name,
            f"Unsupported backend '{backend}'. Use 'numpy' or 'polars'.",
        )

    def is_numpy_async(self) -> bool:
        if self.numpy_impl is None:
            return False
        if inspect.iscoroutinefunction(self.numpy_impl):
            return True
        call_method = getattr(self.numpy_impl, "__call__", None)
        return inspect.iscoroutinefunction(call_method)



from __future__ import annotations

import hashlib
import json
import os
import re
from collections.abc import Mapping
from pathlib import Path
from threading import RLock
from typing import Any

from .errors import OperatorNotFound
from .io_utils import atomic_write_bytes
from .logging import get_logger
from .name_utils import (
    FULL_NAME_SEPARATOR as _FULL_NAME_SEPARATOR,
    namespace_from_source_path,
    normalize_full_name as normalize_common_full_name,
    normalize_namespace as normalize_common_namespace,
    try_split_full_name as try_split_common_full_name,
)
from .namespace_policy import (
    ensure_user_namespace_allowed,
    is_protected_namespace,
    split_full_name as split_policy_full_name,
)
from .core.registry import OperasRegistry

_STORE_LOCK = RLock()
_ENV_STORE_PATH = "JARVIS_OPERAS_PERSIST_FILE"
_ENV_SOURCES_PATH = "JARVIS_OPERAS_SOURCES_FILE"
_ENV_OVERRIDES_PATH = "JARVIS_OPERAS_OVERRIDES_FILE"
_ENV_USER_SOURCE_DIR = "JARVIS_OPERAS_USER_SOURCE_DIR"


def _legacy_env_base_path() -> Path | None:
    raw = os.getenv(_ENV_STORE_PATH)
    if not raw:
        return None
    return Path(raw).expanduser()


def _resolve_store_file_path(
    *,
    explicit_env: str,
    legacy_filename: str,
    default_filename: str,
) -> Path:
    raw = os.getenv(explicit_env)
    if raw:
        return Path(raw).expanduser()

    legacy_path = _legacy_env_base_path()
    if legacy_path is not None:
        if legacy_path.suffix:
            return legacy_path.with_name(legacy_filename)
        return legacy_path / legacy_filename

    return Path.home() / ".jarvis-operas" / default_filename


def get_persist_store_path() -> Path:
    """Return legacy combined persistence file path for compatibility."""

    legacy_path = _legacy_env_base_path()
    if legacy_path is not None:
        return legacy_path
    return Path.home() / ".jarvis-operas" / "user_ops.json"


def get_sources_store_path() -> Path:
    """Return source entries persistence file path."""

    return _resolve_store_file_path(
        explicit_env=_ENV_SOURCES_PATH,
        legacy_filename="sources.json",
        default_filename="sources.json",
    )


def get_overrides_store_path() -> Path:
    """Return override rules persistence file path."""

    return _resolve_store_file_path(
        explicit_env=_ENV_OVERRIDES_PATH,
        legacy_filename="overrides.json",
        default_filename="overrides.json",
    )


def get_user_source_store_dir(*, store_path: Path | None = None) -> Path:
    """Return managed snapshot directory for user operator source files."""

    raw = os.getenv(_ENV_USER_SOURCE_DIR)
    if raw:
        return Path(raw).expanduser()
    base_store = (store_path or get_sources_store_path()).expanduser()
    return base_store.parent / "user_sources"


def persist_user_ops(
    path: str,
    *,
    namespace: str | None = None,
    logger: Any | None = None,
) -> dict[str, Any]:
    """Persist a user operator script path so future processes auto-load it."""

    local_logger = get_logger(logger, action="persist_user_ops")
    store_path = get_sources_store_path()
    resolved_source_path = Path(path).expanduser().resolve()
    requested_namespace = namespace.strip() if isinstance(namespace, str) and namespace.strip() else None
    effective_namespace = ensure_user_namespace_allowed(
        requested_namespace or namespace_from_source_path(resolved_source_path),
        action="persisting user operators",
    )

    snapshot = _snapshot_user_source(
        resolved_source_path,
        store_path=store_path,
    )
    entry: dict[str, Any] = {
        "path": str(resolved_source_path),
        "load_path": snapshot["load_path"],
        "source_hash": snapshot["source_hash"],
        "namespace": effective_namespace,
    }

    with _STORE_LOCK:
        store = _read_sources_store(store_path, local_logger)
        entries = store["entries"]
        matched_index = next(
            (
                idx
                for idx, item in enumerate(entries)
                if item.get("path") == entry["path"]
                and item.get("namespace") == entry.get("namespace")
            ),
            None,
        )
        created = matched_index is None
        updated = False

        if created:
            entries.append(entry)
            _write_sources_store(store_path, store)
        else:
            existing_entry = entries[matched_index]
            changed = any(existing_entry.get(k) != v for k, v in entry.items())
            if changed:
                entries[matched_index] = {**existing_entry, **entry}
                _write_sources_store(store_path, store)
                updated = True

    local_logger.info(
        "persisted user operator source: {} (snapshot: {})",
        str(resolved_source_path),
        entry["load_path"],
    )
    return {
        "entry": dict(entry),
        "store_path": str(store_path),
        "created": created,
        "updated": updated,
    }


def list_persisted_user_ops(*, logger: Any | None = None) -> list[dict[str, Any]]:
    """Return persisted user operator script entries."""

    local_logger = get_logger(logger, action="list_persisted_user_ops")
    store = _read_sources_store(get_sources_store_path(), local_logger)
    return [dict(entry) for entry in store["entries"]]


def delete_persisted_function(
    full_name: str,
    *,
    logger: Any | None = None,
) -> dict[str, Any]:
    """Persist a function deletion override by full operator name."""

    normalized = _normalize_full_name(full_name)
    namespace, _ = split_policy_full_name(normalized)
    ensure_user_namespace_allowed(
        namespace,
        action="function deletion override",
    )
    local_logger = get_logger(logger, action="delete_persisted_function")
    store_path = get_overrides_store_path()

    with _STORE_LOCK:
        store = _read_overrides_store(store_path, local_logger)
        overrides = store["overrides"]
        deleted: list[str] = overrides["deleted_functions"]
        created = normalized not in deleted
        if created:
            deleted.append(normalized)
        overrides["renamed_functions"].pop(normalized, None)
        _write_overrides_store(store_path, store)

    local_logger.info("persisted function deletion override: {}", normalized)
    return {"full_name": normalized, "created": created, "store_path": str(store_path)}


def clear_persisted_function_overrides(
    full_name: str,
    *,
    logger: Any | None = None,
) -> dict[str, Any]:
    """Clear function-level override rules for a specific operator."""

    normalized = _normalize_full_name(full_name)
    local_logger = get_logger(logger, action="clear_persisted_function_overrides")
    store_path = get_overrides_store_path()

    with _STORE_LOCK:
        store = _read_overrides_store(store_path, local_logger)
        overrides = store["overrides"]
        deleted_functions: list[str] = overrides["deleted_functions"]
        renamed_functions: dict[str, str] = overrides["renamed_functions"]

        before_deleted = len(deleted_functions)
        before_renamed = len(renamed_functions)

        overrides["deleted_functions"] = [item for item in deleted_functions if item != normalized]
        overrides["renamed_functions"] = {
            source: target
            for source, target in renamed_functions.items()
            if source != normalized and target != normalized
        }
        _write_overrides_store(store_path, store)

    local_logger.info("cleared function overrides for {}", normalized)
    return {
        "full_name": normalized,
        "removed_deleted": before_deleted - len(store["overrides"]["deleted_functions"]),
        "removed_renamed": before_renamed - len(store["overrides"]["renamed_functions"]),
        "store_path": str(store_path),
    }


def update_persisted_function(
    full_name: str,
    new_full_name: str,
    *,
    logger: Any | None = None,
) -> dict[str, Any]:
    """Persist a function rename/update override."""

    source = _normalize_full_name(full_name)
    target = _normalize_full_name(new_full_name)
    source_namespace, _ = split_policy_full_name(source)
    target_namespace, _ = split_policy_full_name(target)
    ensure_user_namespace_allowed(
        source_namespace,
        action="function rename override source",
    )
    ensure_user_namespace_allowed(
        target_namespace,
        action="function rename override target",
    )
    if source == target:
        raise ValueError("full_name and new_full_name must be different")

    local_logger = get_logger(logger, action="update_persisted_function")
    store_path = get_overrides_store_path()

    with _STORE_LOCK:
        store = _read_overrides_store(store_path, local_logger)
        overrides = store["overrides"]
        overrides["renamed_functions"][source] = target
        overrides["deleted_functions"] = [
            item
            for item in overrides["deleted_functions"]
            if item not in (source, target)
        ]
        _write_overrides_store(store_path, store)

    local_logger.info("persisted function rename override: {} -> {}", source, target)
    return {
        "from": source,
        "to": target,
        "store_path": str(store_path),
    }


def delete_persisted_namespace(
    namespace: str,
    *,
    logger: Any | None = None,
) -> dict[str, Any]:
    """Persist a namespace deletion override."""

    normalized = normalize_common_namespace(namespace)
    ensure_user_namespace_allowed(
        normalized,
        action="namespace deletion override",
    )
    local_logger = get_logger(logger, action="delete_persisted_namespace")
    store_path = get_overrides_store_path()

    with _STORE_LOCK:
        store = _read_overrides_store(store_path, local_logger)
        overrides = store["overrides"]
        deleted: list[str] = overrides["deleted_namespaces"]
        created = normalized not in deleted
        if created:
            deleted.append(normalized)
        overrides["renamed_namespaces"].pop(normalized, None)
        _write_overrides_store(store_path, store)

    local_logger.info("persisted namespace deletion override: {}", normalized)
    return {"namespace": normalized, "created": created, "store_path": str(store_path)}


def clear_persisted_namespace_overrides(
    namespace: str,
    *,
    logger: Any | None = None,
) -> dict[str, Any]:
    """Clear namespace-level (and related function-level) override rules."""

    normalized = normalize_common_namespace(namespace)
    local_logger = get_logger(logger, action="clear_persisted_namespace_overrides")
    store_path = get_overrides_store_path()

    with _STORE_LOCK:
        store = _read_overrides_store(store_path, local_logger)
        overrides = store["overrides"]

        deleted_namespaces: list[str] = overrides["deleted_namespaces"]
        renamed_namespaces: dict[str, str] = overrides["renamed_namespaces"]
        deleted_functions: list[str] = overrides["deleted_functions"]
        renamed_functions: dict[str, str] = overrides["renamed_functions"]

        before_deleted_ns = len(deleted_namespaces)
        before_renamed_ns = len(renamed_namespaces)
        before_deleted_fn = len(deleted_functions)
        before_renamed_fn = len(renamed_functions)

        overrides["deleted_namespaces"] = [
            item for item in deleted_namespaces if item != normalized
        ]
        overrides["renamed_namespaces"] = {
            source: target
            for source, target in renamed_namespaces.items()
            if source != normalized and target != normalized
        }
        overrides["deleted_functions"] = [
            item
            for item in deleted_functions
            if _namespace_from_maybe_full_name(item) != normalized
        ]
        overrides["renamed_functions"] = {
            source: target
            for source, target in renamed_functions.items()
            if _namespace_from_maybe_full_name(source) != normalized
            and _namespace_from_maybe_full_name(target) != normalized
        }
        _write_overrides_store(store_path, store)

    local_logger.info("cleared namespace overrides for {}", normalized)
    return {
        "namespace": normalized,
        "removed_deleted_namespaces": before_deleted_ns - len(store["overrides"]["deleted_namespaces"]),
        "removed_renamed_namespaces": before_renamed_ns - len(store["overrides"]["renamed_namespaces"]),
        "removed_deleted_functions": before_deleted_fn - len(store["overrides"]["deleted_functions"]),
        "removed_renamed_functions": before_renamed_fn - len(store["overrides"]["renamed_functions"]),
        "store_path": str(store_path),
    }


def update_persisted_namespace(
    namespace: str,
    new_namespace: str,
    *,
    logger: Any | None = None,
) -> dict[str, Any]:
    """Persist a namespace rename/update override."""

    source = normalize_common_namespace(namespace)
    target = normalize_common_namespace(new_namespace)
    ensure_user_namespace_allowed(
        source,
        action="namespace rename override source",
    )
    ensure_user_namespace_allowed(
        target,
        action="namespace rename override target",
    )
    if source == target:
        raise ValueError("namespace and new_namespace must be different")

    local_logger = get_logger(logger, action="update_persisted_namespace")
    store_path = get_overrides_store_path()

    with _STORE_LOCK:
        store = _read_overrides_store(store_path, local_logger)
        overrides = store["overrides"]
        overrides["renamed_namespaces"][source] = target
        overrides["deleted_namespaces"] = [
            item for item in overrides["deleted_namespaces"] if item not in (source, target)
        ]
        _write_overrides_store(store_path, store)

    local_logger.info("persisted namespace rename override: {} -> {}", source, target)
    return {"from": source, "to": target, "store_path": str(store_path)}


def apply_persisted_overrides(
    registry: OperasRegistry,
    *,
    logger: Any | None = None,
    strict: bool = False,
) -> dict[str, list[str]]:
    """Apply persisted delete/update overrides for functions and namespaces."""

    local_logger = get_logger(logger, action="apply_persisted_overrides")
    store = _read_overrides_store(get_overrides_store_path(), local_logger)
    overrides = store["overrides"]

    renamed_namespaces = dict(overrides["renamed_namespaces"])
    renamed_functions = dict(overrides["renamed_functions"])
    deleted_namespaces = list(overrides["deleted_namespaces"])
    deleted_functions = list(overrides["deleted_functions"])

    summary = {
        "updated_namespaces": [],
        "updated_functions": [],
        "deleted_namespaces": [],
        "deleted_functions": [],
    }

    for source_ns, target_ns in renamed_namespaces.items():
        if source_ns == target_ns:
            continue
        if is_protected_namespace(source_ns) or is_protected_namespace(target_ns):
            local_logger.debug(
                "skip protected namespace rename override {} -> {}",
                source_ns,
                target_ns,
            )
            continue
        try:
            updated = registry.rename_namespace(source_ns, target_ns)
            if updated:
                summary["updated_namespaces"].extend(updated)
        except Exception as exc:
            if strict:
                raise
            local_logger.warning(
                "skip namespace rename override {} -> {}: {}",
                source_ns,
                target_ns,
                exc,
            )

    for source_full, target_full in renamed_functions.items():
        source_effective = _remap_full_name_namespace(source_full, renamed_namespaces)
        target_effective = _remap_full_name_namespace(target_full, renamed_namespaces)
        if _is_protected_full_name(source_effective) or _is_protected_full_name(target_effective):
            local_logger.debug(
                "skip protected function rename override {} -> {}",
                source_full,
                target_full,
            )
            continue
        try:
            updated_name = registry.rename(source_effective, new_full_name=target_effective)
            summary["updated_functions"].append(updated_name)
        except OperatorNotFound:
            continue
        except Exception as exc:
            if strict:
                raise
            local_logger.warning(
                "skip function rename override {} -> {}: {}",
                source_full,
                target_full,
                exc,
            )

    for namespace in deleted_namespaces:
        namespace_effective = renamed_namespaces.get(namespace, namespace)
        if is_protected_namespace(namespace_effective):
            local_logger.debug(
                "skip protected namespace deletion override {}",
                namespace_effective,
            )
            continue
        try:
            deleted = registry.delete_namespace(namespace_effective)
            if deleted:
                summary["deleted_namespaces"].extend(deleted)
        except Exception as exc:
            if strict:
                raise
            local_logger.warning(
                "skip namespace deletion override {}: {}",
                namespace,
                exc,
            )

    for full_name in deleted_functions:
        candidates = [
            full_name,
            _remap_full_name_namespace(full_name, renamed_namespaces),
            renamed_functions.get(full_name),
        ]
        candidates = [c for c in candidates if c]
        deleted_any = False
        for candidate in dict.fromkeys(candidates):
            if _is_protected_full_name(candidate):
                local_logger.debug(
                    "skip protected function deletion override {}",
                    candidate,
                )
                continue
            try:
                registry.delete(candidate)
                summary["deleted_functions"].append(candidate)
                deleted_any = True
                break
            except OperatorNotFound:
                continue
            except Exception as exc:
                if strict:
                    raise
                local_logger.warning(
                    "skip function deletion override {} (candidate {}): {}",
                    full_name,
                    candidate,
                    exc,
                )
                break
        if not deleted_any:
            continue

    return {
        key: list(dict.fromkeys(values))
        for key, values in summary.items()
    }


def load_persisted_user_ops(
    registry: OperasRegistry,
    *,
    logger: Any | None = None,
    strict: bool = False,
    refresh_sympy: bool = True,
) -> list[str]:
    """Load all persisted user operator scripts into the provided registry."""

    local_logger = get_logger(logger, action="load_persisted_user_ops")
    entries = list_persisted_user_ops(logger=local_logger)
    if not entries:
        return []

    # Import lazily to avoid module cycle during API bootstrap.
    from .loading import load_user_ops

    loaded: list[str] = []
    for entry in entries:
        source_path = entry["path"]
        load_path = entry.get("load_path")
        primary_path = str(load_path).strip() if isinstance(load_path, str) and load_path.strip() else source_path
        namespace = entry.get("namespace")
        candidate_paths = [primary_path]
        if source_path not in candidate_paths:
            candidate_paths.append(source_path)

        selected_path: str | None = None
        for candidate in candidate_paths:
            candidate_path = Path(candidate).expanduser()
            if candidate_path.exists() and candidate_path.is_file():
                selected_path = str(candidate_path)
                break

        if selected_path is None:
            local_logger.warning(
                "skip missing persisted operator file: source={} snapshot={}",
                source_path,
                primary_path,
            )
            continue
        try:
            loaded.extend(
                load_user_ops(
                    selected_path,
                    registry,
                    namespace=namespace,
                    refresh_sympy=refresh_sympy,
                    logger=local_logger,
                )
            )
        except Exception as exc:
            if strict:
                raise
            local_logger.warning("skip invalid persisted operator file '{}': {}", selected_path, exc)

    return list(dict.fromkeys(loaded))


def _empty_sources_store() -> dict[str, Any]:
    return {
        "version": 2,
        "entries": [],
    }


def _empty_overrides_store() -> dict[str, Any]:
    return {
        "version": 2,
        "overrides": {
            "deleted_functions": [],
            "renamed_functions": {},
            "deleted_namespaces": [],
            "renamed_namespaces": {},
        },
    }


def _normalize_full_name(full_name: str) -> str:
    return normalize_common_full_name(full_name)


def _is_protected_full_name(full_name: str) -> bool:
    try:
        namespace, _ = split_policy_full_name(full_name)
    except Exception:
        return False
    return is_protected_namespace(namespace)


def _namespace_from_maybe_full_name(value: str) -> str:
    split = try_split_common_full_name(value)
    if split is None:
        return ""
    namespace, _ = split
    return namespace


def _remap_full_name_namespace(
    full_name: str,
    renamed_namespaces: Mapping[str, str],
) -> str:
    normalized = _normalize_full_name(full_name)
    namespace, short_name = normalized.split(_FULL_NAME_SEPARATOR, 1)
    mapped_namespace = renamed_namespaces.get(namespace, namespace)
    return f"{mapped_namespace}{_FULL_NAME_SEPARATOR}{short_name}"


def _snapshot_user_source(
    source_path: Path,
    *,
    store_path: Path,
) -> dict[str, str]:
    if not source_path.exists() or not source_path.is_file():
        raise FileNotFoundError(f"user operator source does not exist: {source_path}")

    payload = source_path.read_bytes()
    source_hash = hashlib.sha256(payload).hexdigest()
    snapshot_dir = get_user_source_store_dir(store_path=store_path)
    safe_stem = _sanitize_snapshot_stem(source_path.stem)
    suffix = source_path.suffix or ".py"
    snapshot_path = snapshot_dir / f"{safe_stem}-{source_hash[:16]}{suffix}"

    if not snapshot_path.exists():
        atomic_write_bytes(snapshot_path, payload)

    return {
        "load_path": str(snapshot_path.resolve()),
        "source_hash": source_hash,
    }


def _sanitize_snapshot_stem(stem: str) -> str:
    normalized = re.sub(r"[^0-9A-Za-z_-]+", "_", stem).strip("_")
    if not normalized:
        return "user_ops"
    return normalized


def _legacy_combined_store_path() -> Path:
    return get_persist_store_path().expanduser()


def _read_store_mapping_with_legacy_fallback(
    path: Path,
    logger: Any,
    *,
    label: str,
) -> Mapping[str, Any] | None:
    raw = _read_json_mapping(path, logger, label=label)
    if raw is not None:
        return raw

    legacy_path = _legacy_combined_store_path()
    if legacy_path == path:
        return None
    return _read_json_mapping(
        legacy_path,
        logger,
        label="legacy persist store",
    )


def _read_sources_store(path: Path, logger: Any) -> dict[str, Any]:
    raw = _read_store_mapping_with_legacy_fallback(
        path,
        logger,
        label="sources persist store",
    )
    if raw is None:
        return _empty_sources_store()
    entries = _normalize_persist_entries(raw.get("entries", []))
    return {
        "version": 2,
        "entries": entries,
    }


def _read_overrides_store(path: Path, logger: Any) -> dict[str, Any]:
    raw = _read_store_mapping_with_legacy_fallback(
        path,
        logger,
        label="overrides persist store",
    )
    if raw is None:
        return _empty_overrides_store()

    raw_overrides = raw.get("overrides")
    if raw_overrides is None:
        raw_overrides = raw
    if not isinstance(raw_overrides, Mapping):
        raw_overrides = {}

    return {
        "version": 2,
        "overrides": _normalize_override_rules(raw_overrides),
    }


def _normalize_persist_entries(raw_entries: Any) -> list[dict[str, Any]]:
    if not isinstance(raw_entries, list):
        return []

    entries: list[dict[str, Any]] = []
    for item in raw_entries:
        if not isinstance(item, Mapping):
            continue
        item_path = item.get("path")
        if not isinstance(item_path, str) or not item_path.strip():
            continue
        normalized: dict[str, Any] = {"path": item_path.strip()}

        item_namespace = item.get("namespace")
        if isinstance(item_namespace, str) and item_namespace.strip():
            normalized["namespace"] = item_namespace.strip()

        item_load_path = item.get("load_path")
        if isinstance(item_load_path, str) and item_load_path.strip():
            normalized["load_path"] = item_load_path.strip()

        item_source_hash = item.get("source_hash")
        if isinstance(item_source_hash, str) and item_source_hash.strip():
            normalized["source_hash"] = item_source_hash.strip()
        entries.append(normalized)

    return entries


def _normalize_override_rules(raw_overrides: Mapping[str, Any]) -> dict[str, Any]:
    raw_deleted_functions = raw_overrides.get("deleted_functions", [])
    if not isinstance(raw_deleted_functions, list):
        raw_deleted_functions = []

    raw_renamed_functions = raw_overrides.get("renamed_functions", {})
    if not isinstance(raw_renamed_functions, Mapping):
        raw_renamed_functions = {}

    raw_deleted_namespaces = raw_overrides.get("deleted_namespaces", [])
    if not isinstance(raw_deleted_namespaces, list):
        raw_deleted_namespaces = []

    raw_renamed_namespaces = raw_overrides.get("renamed_namespaces", {})
    if not isinstance(raw_renamed_namespaces, Mapping):
        raw_renamed_namespaces = {}

    deleted_functions: list[str] = []
    for item in raw_deleted_functions:
        if not isinstance(item, str):
            continue
        try:
            deleted_functions.append(_normalize_full_name(item))
        except ValueError:
            continue

    renamed_functions: dict[str, str] = {}
    for source, target in raw_renamed_functions.items():
        if not isinstance(source, str) or not isinstance(target, str):
            continue
        try:
            source_normalized = _normalize_full_name(source)
            target_normalized = _normalize_full_name(target)
        except ValueError:
            continue
        renamed_functions[source_normalized] = target_normalized

    deleted_namespaces: list[str] = []
    for item in raw_deleted_namespaces:
        if not isinstance(item, str):
            continue
        try:
            deleted_namespaces.append(normalize_common_namespace(item))
        except ValueError:
            continue

    renamed_namespaces: dict[str, str] = {}
    for source, target in raw_renamed_namespaces.items():
        if not isinstance(source, str) or not isinstance(target, str):
            continue
        try:
            source_normalized = normalize_common_namespace(source)
            target_normalized = normalize_common_namespace(target)
        except ValueError:
            continue
        renamed_namespaces[source_normalized] = target_normalized

    return {
        "deleted_functions": list(dict.fromkeys(deleted_functions)),
        "renamed_functions": renamed_functions,
        "deleted_namespaces": list(dict.fromkeys(deleted_namespaces)),
        "renamed_namespaces": renamed_namespaces,
    }


def _read_json_mapping(path: Path, logger: Any, *, label: str) -> Mapping[str, Any] | None:
    if not path.exists():
        return None
    try:
        raw = json.loads(path.read_text(encoding="utf-8"))
    except Exception as exc:
        logger.warning("invalid {} '{}': {}. ignoring.", label, str(path), exc)
        return None
    if not isinstance(raw, Mapping):
        logger.warning("invalid {} '{}': expected object. ignoring.", label, str(path))
        return None
    return raw


def _write_store_json(path: Path, payload: Mapping[str, Any]) -> None:
    atomic_write_bytes(
        path,
        json.dumps(payload, ensure_ascii=False, indent=2).encode("utf-8"),
    )


def _write_sources_store(path: Path, store: Mapping[str, Any]) -> None:
    entries = _normalize_persist_entries(store.get("entries", []))
    payload = {
        "version": 2,
        "entries": [
            {
                "path": entry["path"],
                **({"namespace": entry["namespace"]} if "namespace" in entry else {}),
                **({"load_path": entry["load_path"]} if "load_path" in entry else {}),
                **({"source_hash": entry["source_hash"]} if "source_hash" in entry else {}),
            }
            for entry in entries
        ],
    }
    _write_store_json(path, payload)


def _write_overrides_store(path: Path, store: Mapping[str, Any]) -> None:
    raw_overrides = store.get("overrides", {})
    if not isinstance(raw_overrides, Mapping):
        raw_overrides = {}
    payload = {
        "version": 2,
        "overrides": _normalize_override_rules(raw_overrides),
    }
    _write_store_json(path, payload)



#!/usr/bin/env python3
from __future__ import annotations

from pathlib import Path

import sympy as sp
from sympy.utilities.lambdify import lambdify

import jarvis_operas.api as api_mod
import jarvis_operas.integration as integration
from jarvis_operas import OperaFunction, OperasRegistry
from jarvis_operas.integration import (
    build_register_dicts,
    build_sympy_dicts,
    refresh_sympy_dicts_if_global_registry,
)


def _register(registry: OperasRegistry, namespace: str, name: str, fn) -> None:
    registry.register(
        OperaFunction(
            namespace=namespace,
            name=name,
            arity=None,
            return_dtype=None,
            numpy_impl=fn,
        )
    )


def test_build_register_dicts_returns_full_name_to_callable_only() -> None:
    registry = OperasRegistry()
    _register(registry, "dmdd", "LZSI2024", lambda x: x * 2.0)

    mapping = build_register_dicts(registry)
    assert callable(mapping["dmdd.LZSI2024"])
    assert set(mapping.keys()) == {"dmdd.LZSI2024"}


def test_build_sympy_dicts_for_namespace_function_parsing_and_eval() -> None:
    registry = OperasRegistry()
    _register(registry, "dmdd", "LZSI2024", lambda x: x * 2.0)
    _register(registry, "math", "add", lambda x, y: x + y)

    mapping = build_register_dicts(registry)
    func_locals, numeric_funcs = build_sympy_dicts(mapping, namespaces=["dmdd"])

    expr = sp.sympify("dmdd.LZSI2024(mchi)", locals=func_locals)
    num_expr = lambdify(["mchi"], expr, modules=[numeric_funcs, "numpy"])

    assert float(num_expr(95.0)) == 190.0
    assert "math" not in func_locals


def test_refresh_sympy_dicts_if_global_registry_updates_module_level_dicts(monkeypatch) -> None:
    sentinel_registry = object()
    monkeypatch.setattr(api_mod, "_global_operas_registry", sentinel_registry)
    monkeypatch.setattr(
        integration,
        "build_register_dicts",
        lambda registry=None: {"dmdd.LZSI2024": lambda x: x + 1.0},
    )

    old_func_locals = integration.func_locals
    old_numeric_funcs = integration.numeric_funcs
    refreshed = refresh_sympy_dicts_if_global_registry(sentinel_registry)

    expr = sp.sympify("dmdd.LZSI2024(mchi)", locals=integration.func_locals)
    num_expr = lambdify(["mchi"], expr, modules=[integration.numeric_funcs, "numpy"])
    assert refreshed is True
    assert old_func_locals is integration.func_locals
    assert old_numeric_funcs is integration.numeric_funcs
    assert float(num_expr(2.0)) == 3.0


def test_refresh_sympy_dicts_if_global_registry_skips_non_global_registry(monkeypatch) -> None:
    sentinel_registry = object()
    monkeypatch.setattr(api_mod, "_global_operas_registry", sentinel_registry)
    assert refresh_sympy_dicts_if_global_registry(object()) is False


def test_integration_does_not_access_api_private_global_symbol() -> None:
    repo_root = Path(__file__).resolve().parent.parent
    body = (repo_root / "jarvis_operas" / "integration.py").read_text(encoding="utf-8")
    assert "._global_operas_registry" not in body
    assert 'getattr(_api, "_global_operas_registry"' not in body
